# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

import typing_extensions as ty_ext

from keystoneauth1 import discover

if ty.TYPE_CHECKING:
    from keystoneauth1.access import access
    from keystoneauth1 import session as ks_session

# NOTE(jamielennox): The AUTH_INTERFACE is a special value that can be
# requested from get_endpoint. If a plugin receives this as the value of
# 'interface' it should return the initial URL that was passed to the plugin.
AUTH_INTERFACE = object()

IDENTITY_AUTH_HEADER_NAME = 'X-Auth-Token'


BaseAuthPluginT = ty.TypeVar(
    'BaseAuthPluginT', bound='BaseAuthPlugin', covariant=True
)


class ConnectionParams(ty.TypedDict):
    # https://github.com/python/typeshed/blob/24c78b9e0/stubs/requests/requests/sessions.pyi#L82
    cert: ty_ext.NotRequired[str | tuple[str, str] | None]
    # https://github.com/python/typeshed/blob/24c78b9e0/stubs/requests/requests/sessions.pyi#L108
    verify: ty_ext.NotRequired[bool | str | None]


class BaseAuthPlugin:
    """The basic structure of an authentication plugin.

    .. note::
        See :doc:`/authentication-plugins` for a description of plugins
        provided by this library.

    """

    def __init__(self) -> None:
        self._discovery_cache: dict[str, discover.Discover] = {}

    def get_token(self, session: 'ks_session.Session') -> str | None:
        """Obtain a token.

        How the token is obtained is up to the plugin. If it is still valid
        it may be re-used, retrieved from cache or invoke an authentication
        request against a server.

        Returning None will indicate that no token was able to be retrieved.

        This function is misplaced as it should only be required for auth
        plugins that use the 'X-Auth-Token' header. However due to the way
        plugins evolved this method is required and often called to trigger an
        authentication request on a new plugin.

        When implementing a new plugin it is advised that you implement this
        method, however if you don't require the 'X-Auth-Token' header override
        the `get_headers` method instead.

        :param session: A session object so the plugin can make HTTP calls.
        :type session: keystoneauth1.session.Session

        :return: A token to use.
        :rtype: string
        """
        return None

    def get_auth_ref(
        self, session: 'ks_session.Session'
    ) -> ty.Optional['access.AccessInfo']:
        """Return the authentication reference of an auth plugin.

        :param session: A session object to be used for communication
        :type session: keystoneauth1.session.session

        :return: An access info object if supported.
        :rtype: `keystoneauth1.access.AccessInfo` or None
        """
        return None

    def get_headers(
        self, session: 'ks_session.Session'
    ) -> dict[str, str] | None:
        """Fetch authentication headers for message.

        This is a more generalized replacement of the older get_token to allow
        plugins to specify different or additional authentication headers to
        the OpenStack standard 'X-Auth-Token' header.

        How the authentication headers are obtained is up to the plugin. If the
        headers are still valid they may be re-used, retrieved from cache or
        the plugin may invoke an authentication request against a server.

        The default implementation of get_headers calls the `get_token` method
        to enable older style plugins to continue functioning unchanged.
        Subclasses should feel free to completely override this function to
        provide the headers that they want.

        Returning None will indicate that no token was able to be retrieved and
        that authorization was a failure. Adding no authentication data can be
        achieved by returning an empty dictionary.

        :param session: The session object that the auth_plugin belongs to.
        :type session: keystoneauth1.session.Session

        :returns: Headers that are set to authenticate a message or None for
                  failure. Note that when checking this value that the empty
                  dict is a valid, non-failure response.
        :rtype: dict
        """
        token = self.get_token(session)

        if not token:
            return None

        return {IDENTITY_AUTH_HEADER_NAME: token}

    def get_endpoint_data(
        self,
        session: 'ks_session.Session',
        *,
        endpoint_override: str | None = None,
        discover_versions: bool = True,
        **kwargs: ty.Any,
    ) -> discover.EndpointData | None:
        """Return a valid endpoint data for a service.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param str endpoint_override: URL to use for version discovery.
        :param bool discover_versions: Whether to get version metadata from
                                       the version discovery document even
                                       if it major api version info can be
                                       inferred from the url.
                                       (optional, defaults to True)
        :param kwargs: Ignored.

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: Valid EndpointData or None if not available.
        :rtype: `keystoneauth1.discover.EndpointData` or None
        """
        if not endpoint_override:
            return None
        endpoint_data = discover.EndpointData(catalog_url=endpoint_override)

        if endpoint_data.api_version and not discover_versions:
            return endpoint_data

        return endpoint_data.get_versioned_data(
            session,
            cache=self._discovery_cache,
            discover_versions=discover_versions,
        )

    def get_api_major_version(
        self,
        session: 'ks_session.Session',
        *,
        endpoint_override: str | None = None,
        **kwargs: ty.Any,
    ) -> tuple[int | float, ...] | None:
        """Get the major API version from the endpoint.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param str endpoint_override: URL to use for version discovery.
        :param kwargs: Ignored.

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: Valid EndpointData or None if not available.
        :rtype: `keystoneauth1.discover.EndpointData` or None
        """
        endpoint_data = self.get_endpoint_data(
            session,
            endpoint_override=endpoint_override,
            discover_versions=False,
            **kwargs,
        )
        if endpoint_data is None:
            return None

        if endpoint_data.api_version is None:
            # No version detected from the URL, trying full discovery.
            endpoint_data = self.get_endpoint_data(
                session,
                endpoint_override=endpoint_override,
                discover_versions=True,
                **kwargs,
            )

        if endpoint_data and endpoint_data.api_version:
            return endpoint_data.api_version

        return None

    def get_all_version_data(
        self,
        session: 'ks_session.Session',
        interface: str | list[str] | None = 'public',
        region_name: str | None = None,
        service_type: str | None = None,
    ) -> dict[str, dict[str, dict[str, list[discover.VersionData]]]]:
        """Get version data for all services in the catalog.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param interface:
            Type of endpoint to get version data for. Can be a single value
            or a list of values. A value of None indicates that all interfaces
            should be queried. (optional, defaults to public)
        :param string region_name:
            Region of endpoints to get version data for. A valueof None
            indicates that all regions should be queried. (optional, defaults
            to None)
        :param string service_type:
            Limit the version data to a single service. (optional, defaults
            to None)

        :returns: A dictionary keyed by region_name with values containing
            dictionaries keyed by interface with values being a list of
            :class:`~keystoneauth1.discover.VersionData`.
        """
        raise NotImplementedError()

    def get_endpoint(
        self, session: 'ks_session.Session', **kwargs: ty.Any
    ) -> str | None:
        """Return an endpoint for the client.

        There are no required keyword arguments to ``get_endpoint`` as a plugin
        implementation should use best effort with the information available to
        determine the endpoint. However there are certain standard options that
        will be generated by the clients and should be used by plugins:

        - ``service_type``: what sort of service is required.
        - ``service_name``: the name of the service in the catalog.
        - ``interface``: what visibility the endpoint should have.
        - ``region_name``: the region the endpoint exists in.

        :param session: The session object that the auth_plugin belongs to.
        :type session: keystoneauth1.session.Session
        :param kwargs: Ignored.

        :returns: The base URL that will be used to talk to the required
                  service or None if not available.
        :rtype: string
        """
        endpoint_data = self.get_endpoint_data(
            session, discover_versions=False, **kwargs
        )
        if not endpoint_data:
            return None
        return endpoint_data.url

    def get_connection_params(
        self, session: 'ks_session.Session'
    ) -> ConnectionParams:
        """Return any additional connection parameters required for the plugin.

        :param session: The session object that the auth_plugin belongs to.
        :type session: keystoneauth1.session.Session

        :returns: Parameters that are passed to the requests library. Only the
                  ``cert`` and ``verify`` parameters may be returned.
        :rtype: dict
        """
        return {}

    def invalidate(self) -> bool:
        """Invalidate the current authentication data.

        This should result in fetching a new token on next call.

        A plugin may be invalidated if an Unauthorized HTTP response is
        returned to indicate that the token may have been revoked or is
        otherwise now invalid.

        :returns: True if there was something that the plugin did to
                  invalidate. This means that it makes sense to try again. If
                  nothing happens returns False to indicate give up.
        :rtype: bool
        """
        return False

    def get_user_id(self, session: 'ks_session.Session') -> str | None:
        """Return a unique user identifier of the plugin.

        Wherever possible the user id should be inferred from the token however
        there are certain URLs and other places that require access to the
        currently authenticated user id.

        :param session: A session object so the plugin can make HTTP calls.
        :type session: keystoneauth1.session.Session

        :returns: A user identifier or None if one is not available.
        :rtype: str
        """
        return None

    def get_project_id(self, session: 'ks_session.Session') -> str | None:
        """Return the project id that we are authenticated to.

        Wherever possible the project id should be inferred from the token
        however there are certain URLs and other places that require access to
        the currently authenticated project id.

        :param session: A session object so the plugin can make HTTP calls.
        :type session: keystoneauth1.session.Session

        :returns: A project identifier or None if one is not available.
        :rtype: str
        """
        return None

    def get_sp_auth_url(
        self, session: 'ks_session.Session', sp_id: str
    ) -> str | None:
        """Return auth_url from the Service Provider object.

        This url is used for obtaining unscoped federated token from remote
        cloud.

        :param sp_id: ID of the Service Provider to be queried.
        :type sp_id: string

        :returns: A Service Provider auth_url or None if one is not available.
        :rtype: str

        """
        return None

    def get_sp_url(
        self, session: 'ks_session.Session', sp_id: str
    ) -> str | None:
        """Return sp_url from the Service Provider object.

        This url is used for passing SAML2 assertion to the remote cloud.

        :param sp_id: ID of the Service Provider to be queried.
        :type sp_id: str

        :returns: A Service Provider sp_url or None if one is not available.
        :rtype: str

        """
        return None

    def get_cache_id(self) -> str | None:
        """Fetch an identifier that uniquely identifies the auth options.

        The returned identifier need not be decomposable or otherwise provide
        anyway to recreate the plugin. It should not contain sensitive data in
        plaintext.

        This string MUST change if any of the parameters that are used to
        uniquely identity this plugin change.

        If get_cache_id returns a str value suggesting that caching is
        supported then get_auth_cache and set_auth_cache must also be
        implemented.

        :returns: A unique string for the set of options
        :rtype: str or None if this is unsupported or unavailable.
        """
        return None

    def get_auth_state(self) -> object:
        """Retrieve the current authentication state for the plugin.

        Retrieve any internal state that represents the authenticated plugin.

        This should not fetch any new data if it is not present.

        :raises NotImplementedError: if the plugin does not support this
            feature.

        :returns: raw python data (which can be JSON serialized) that can be
                  moved into another plugin (of the same type) to have the
                  same authenticated state.
        :rtype: object or None if unauthenticated.
        """
        raise NotImplementedError()

    def set_auth_state(self, data: str) -> None:
        """Install existing authentication state for a plugin.

        Take the output of get_auth_state and install that authentication state
        into the current authentication plugin.

        :raises NotImplementedError: if the plugin does not support this
            feature.
        """
        raise NotImplementedError()


class FixedEndpointPlugin(BaseAuthPlugin):
    """A base class for plugins that have one fixed endpoint."""

    def __init__(self, endpoint: str | None = None):
        super().__init__()
        self.endpoint = endpoint

    def get_endpoint(
        self,
        session: 'ks_session.Session',
        endpoint_override: str | None = None,
        **kwargs: ty.Any,
    ) -> str | None:
        """Return the supplied endpoint.

        Using this plugin the same endpoint is returned regardless of the
        parameters passed to the plugin. endpoint_override overrides the
        endpoint specified when constructing the plugin.
        """
        return endpoint_override or self.endpoint

    def get_endpoint_data(
        self,
        session: 'ks_session.Session',
        *,
        endpoint_override: str | None = None,
        discover_versions: bool = True,
        **kwargs: ty.Any,
    ) -> discover.EndpointData | None:
        """Return a valid endpoint data for a the service.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param str endpoint_override: URL to use for version discovery.
        :param bool discover_versions: Whether to get version metadata from
                                       the version discovery document even
                                       if it major api version info can be
                                       inferred from the url.
                                       (optional, defaults to True)
        :param kwargs: Ignored.

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: Valid EndpointData or None if not available.
        :rtype: `keystoneauth1.discover.EndpointData` or None
        """
        return super().get_endpoint_data(
            session,
            endpoint_override=endpoint_override or self.endpoint,
            discover_versions=discover_versions,
            **kwargs,
        )



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from lxml import etree
from testtools import matchers


class XMLEquals:
    """Parses two XML documents from strings and compares the results."""

    def __init__(self, expected):
        self.expected = expected

    def __str__(self):
        """Return string representation of xml document info."""
        return f"{self.__class__.__name__}({self.expected!r})"

    def match(self, other):
        def xml_element_equals(expected_doc, observed_doc):
            """Test whether two XML documents are equivalent.

            This is a recursive algorithm that operates on each element in
            the hierarchy. Siblings are sorted before being checked to
            account for two semantically equivalent documents where siblings
            appear in different document order.

            The sorting algorithm is a little weak in that it could fail for
            documents where siblings at a given level are the same, but have
            different children.

            """
            if expected_doc.tag != observed_doc.tag:
                return False

            if expected_doc.attrib != observed_doc.attrib:
                return False

            def _sorted_children(doc):
                return sorted(doc.getchildren(), key=lambda el: el.tag)

            expected_children = _sorted_children(expected_doc)
            observed_children = _sorted_children(observed_doc)

            if len(expected_children) != len(observed_children):
                return False

            for expected_el, observed_el in zip(
                expected_children, observed_children
            ):
                if not xml_element_equals(expected_el, observed_el):
                    return False

            return True

        parser = etree.XMLParser(remove_blank_text=True)
        expected_doc = etree.fromstring(self.expected.strip(), parser)
        observed_doc = etree.fromstring(other.strip(), parser)

        if xml_element_equals(expected_doc, observed_doc):
            return

        return XMLMismatch(self.expected, other)


class XMLMismatch(matchers.Mismatch):
    def __init__(self, expected, other):
        self.expected = expected
        self.other = other

    def describe(self):
        def pretty_xml(xml):
            parser = etree.XMLParser(remove_blank_text=True)
            doc = etree.fromstring(xml.strip(), parser)
            return etree.tostring(
                doc, encoding='utf-8', pretty_print=True
            ).decode('utf-8')

        return (
            f'expected =\n{pretty_xml(self.expected)}\n'
            f'actual =\n{pretty_xml(self.other)}'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty
import warnings

from keystoneauth1.identity.v3 import base
from keystoneauth1 import loading
from keystoneauth1 import plugin

__all__ = ('MultiFactor',)


class MultiFactor(base.Auth):
    """A plugin for authenticating with multiple auth methods.

    :param string auth_url: Identity service endpoint for authentication.
    :param string auth_methods: names of the methods to authenticate with.
    :param string trust_id: Trust ID for trust scoping.
    :param string system_scope: System information to scope to.
    :param string domain_id: Domain ID for domain scoping.
    :param string domain_name: Domain name for domain scoping.
    :param string project_id: Project ID for project scoping.
    :param string project_name: Project name for project scoping.
    :param string project_domain_id: Project's domain ID for project.
    :param string project_domain_name: Project's domain name for project.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True

    Also accepts various keyword args based on which methods are specified.
    """

    def __init__(
        self,
        auth_url: str,
        auth_methods: list[str],
        *,
        unscoped: bool = False,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
        **kwargs: ty.Any,
    ):
        method_instances: list[base.AuthMethod] = []
        method_keys: set[str] = set()
        for method in auth_methods:
            # Using the loaders we pull the related auth method class
            loader: loading.BaseLoader[plugin.BaseAuthPlugin] = (
                loading.get_plugin_loader(method)
            )
            plugin_class = loader.plugin_class

            if issubclass(plugin_class, base.AuthConstructor):  # legacy path
                warnings.warn(
                    f"Support for {base.AuthConstructor.__qualname__} is "
                    f"deprecated and will be removed in a future release. "
                    f"Plugins should subclass for {base.Auth.__qualname__}.",
                    category=DeprecationWarning,
                )
                method_class = plugin_class._auth_method_class
                method_parameters = method_class._method_parameters or []
            elif issubclass(plugin_class, base.Auth) and isinstance(
                plugin_class, base.SupportsMultiFactor
            ):
                method_class = plugin_class._auth_method_class
                method_parameters = list(method_class.__annotations__)
            else:
                raise TypeError(
                    'The multifactor auth method can only be used with v3 '
                    'auth plugins that implement the SupportsMultiFactor '
                    'protocol'
                )

            # We build some new kwargs for the method from required parameters
            method_kwargs = {}
            for key in method_parameters:
                method_kwargs[key] = kwargs.get(key, None)
                # we also add them to method_keys to pop later from global
                # kwargs rather than here as other methods may need them too
                method_keys.add(key)

            # We initialize the method class using just required kwargs
            method_instances.append(method_class(**method_kwargs))

        # We now pop all the keys used for methods as otherwise they get passed
        # to the super class and throw errors
        for key in method_keys:
            kwargs.pop(key, None)

        # This should now be empty
        assert not kwargs  # nosec B101

        super().__init__(
            auth_url=auth_url,
            auth_methods=method_instances,
            unscoped=unscoped,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any
import uuid

import fixtures

from keystoneauth1 import discover
from keystoneauth1 import loading
from keystoneauth1 import plugin

__all__ = ('LoadingFixture', 'TestPlugin')


DEFAULT_TEST_ENDPOINT = 'https://openstack.example.com/%(service_type)s'


def _format_endpoint(endpoint: str, **kwargs: Any) -> str:
    # can't format AUTH_INTERFACE object so replace with string
    if kwargs.get('service_type') is plugin.AUTH_INTERFACE:
        kwargs['service_type'] = 'identity'

    version = kwargs.get('version')
    if version:
        discover.normalize_version_number(version)
        kwargs['version'] = ".".join(str(v) for v in version)

    return endpoint % kwargs  # pass kwargs ok?


class TestPlugin(plugin.BaseAuthPlugin):
    """A simple plugin that returns what you gave it for testing.

    When testing services that use authentication plugins you often want to
    stub out the authentication calls and focus on the important part of your
    service. This plugin acts like a real keystoneauth plugin and returns known
    standard values without having to stub out real keystone responses.

    Note that this plugin is a BaseAuthPlugin and not a BaseIdentityPlugin.
    This means it implements the basic plugin interface that services should be
    using but does not implement get_auth_ref. get_auth_ref should not be
    relied upon by services because a user could always configure the service
    to use a non-keystone auth.

    :param str token: The token to include in authenticated requests.
    :param str endpoint: The endpoint to respond to service lookups with.
    :param str user_id: The user_id to report for the authenticated user.
    :param str project_id: The project_id to report for the authenticated user.
    """

    auth_type = 'test_plugin'

    def __init__(
        self,
        token: str | None = None,
        endpoint: str | None = None,
        user_id: str | None = None,
        project_id: str | None = None,
    ) -> None:
        super().__init__()

        self.token = token or uuid.uuid4().hex
        self.endpoint = endpoint or DEFAULT_TEST_ENDPOINT
        self.user_id = user_id or uuid.uuid4().hex
        self.project_id = project_id or uuid.uuid4().hex

    def get_endpoint(self, session: Any, **kwargs: Any) -> str:
        return _format_endpoint(self.endpoint, **kwargs)

    def get_token(self, session: Any) -> str:
        return self.token

    def get_user_id(self, session: Any) -> str:
        return self.user_id

    def get_project_id(self, session: Any) -> str:
        return self.project_id

    def invalidate(self) -> bool:
        self.token = uuid.uuid4().hex
        return True

    # NOTE(jamielennox): You'll notice there's no get_access/get_auth_ref
    # function here. These functions are only part of identity plugins, which
    # whilst the most common are not the only way you can authenticate. Your
    # application should really only rely on the presence of the above
    # functions, everything else is on a best effort basis.
    # TODO(stephenfin): This is no longer true as of change
    # I239dc34fdcbff9bd658dc89ddb1536c7f0df9395 which added a 'get_auth_ref'
    # method.


class _TestPluginLoader(loading.BaseLoader[TestPlugin]):
    def __init__(self, plugin: TestPlugin) -> None:
        super().__init__()
        self._plugin = plugin

    def create_plugin(self, **kwargs: Any) -> TestPlugin:
        return self._plugin

    def get_options(self) -> list[Any]:
        return []


class LoadingFixture(fixtures.Fixture):
    """A fixture that will stub out all plugin loading calls.

    When using keystoneauth plugins loaded from config, CLI or elsewhere it is
    often difficult to handle the plugin parts in tests because we don't have a
    reasonable default.

    This fixture will create a :py:class:`TestPlugin` that will be
    returned for all calls to plugin loading so you can simply bypass the
    authentication steps and return something well known.

    :param str token: The token to include in authenticated requests.
    :param str endpoint: The endpoint to respond to service lookups with.
    :param str user_id: The user_id to report for the authenticated user.
    :param str project_id: The project_id to report for the authenticated user.
    """

    MOCK_POINT = 'keystoneauth1.loading.base.get_plugin_loader'

    def __init__(
        self,
        token: str | None = None,
        endpoint: str | None = None,
        user_id: str | None = None,
        project_id: str | None = None,
    ) -> None:
        super().__init__()

        # these are created and saved here so that a test could use them
        self.token = token or uuid.uuid4().hex
        self.endpoint = endpoint or DEFAULT_TEST_ENDPOINT
        self.user_id = user_id or uuid.uuid4().hex
        self.project_id = project_id or uuid.uuid4().hex

    def setUp(self) -> None:
        super().setUp()

        self.useFixture(
            fixtures.MonkeyPatch(self.MOCK_POINT, self.get_plugin_loader)
        )

    def create_plugin(self) -> TestPlugin:
        return TestPlugin(
            token=self.token,
            endpoint=self.endpoint,
            user_id=self.user_id,
            project_id=self.project_id,
        )

    def get_plugin_loader(self, auth_type: str) -> _TestPluginLoader:
        plugin = self.create_plugin()
        plugin.auth_type = auth_type
        return _TestPluginLoader(plugin)

    def get_endpoint(self, path: str | None = None, **kwargs: Any) -> str:
        """Utility function to get the endpoint the plugin would return.

        This function is provided as a convenience so you can do comparisons in
        your tests. Overriding it will not affect the endpoint returned by the
        plugin.

        :param str path: The path to append to the plugin endpoint.
        """
        endpoint = _format_endpoint(self.endpoint, **kwargs)

        if path:
            endpoint = "{}/{}".format(endpoint.rstrip('/'), path.lstrip('/'))

        return endpoint



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import exceptions
from keystoneauth1.loading import base
from keystoneauth1.loading import opts

if ty.TYPE_CHECKING:
    from keystoneauth1.identity import base as base_plugins
    from keystoneauth1.identity import generic as generic_plugins
    from keystoneauth1.identity import v2 as v2_plugins
    from keystoneauth1.identity import v3 as v3_plugins

__all__ = (
    'BaseIdentityLoader',
    'BaseV2Loader',
    'BaseV3Loader',
    'BaseFederationLoader',
    'BaseGenericLoader',
)


PluginT = ty.TypeVar('PluginT', bound='base_plugins.BaseIdentityPlugin')
V2PluginT = ty.TypeVar('V2PluginT', bound='v2_plugins.Auth')
V3PluginT = ty.TypeVar('V3PluginT', bound='v3_plugins.BaseAuth')
V3FederationPluginT = ty.TypeVar(
    'V3FederationPluginT', bound='v3_plugins.FederationBaseAuth'
)
GenericPluginT = ty.TypeVar(
    'GenericPluginT', bound='generic_plugins.BaseGenericPlugin'
)


class BaseIdentityLoader(base.BaseLoader[PluginT]):
    """Base Option handling for identity plugins.

    This class defines options and handling that should be common across all
    plugins that are developed against the OpenStack identity service. It
    provides the options expected by the
    :py:class:`keystoneauth1.identity.BaseIdentityPlugin` class.
    """

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [opts.Opt('auth-url', required=True, help='Authentication URL')]
        )

        return options


class BaseV2Loader(BaseIdentityLoader[V2PluginT]):
    """Base Option handling for identity plugins.

    This class defines options and handling that should be common to the V2
    identity API. It provides the options expected by the
    :py:class:`keystoneauth1.identity.v2.Auth` class.
    """

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                opts.Opt('tenant-id', help='Tenant ID'),
                opts.Opt('tenant-name', help='Tenant Name'),
                opts.Opt(
                    'trust-id', help='ID of the trust to use as a trustee use'
                ),
            ]
        )

        return options


class BaseV3Loader(BaseIdentityLoader[V3PluginT]):
    """Base Option handling for identity plugins.

    This class defines options and handling that should be common to the V3
    identity API. It provides the options expected by the
    :py:class:`keystoneauth1.identity.v3.Auth` class.
    """

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                opts.Opt('system-scope', help='Scope for system operations'),
                opts.Opt('domain-id', help='Domain ID to scope to'),
                opts.Opt('domain-name', help='Domain name to scope to'),
                opts.Opt('project-id', help='Project ID to scope to'),
                opts.Opt('project-name', help='Project name to scope to'),
                opts.Opt(
                    'project-domain-id', help='Domain ID containing project'
                ),
                opts.Opt(
                    'project-domain-name',
                    help='Domain name containing project',
                ),
                opts.Opt(
                    'trust-id', help='ID of the trust to use as a trustee use'
                ),
            ]
        )

        return options

    def load_from_options(self, **kwargs: ty.Any) -> V3PluginT:
        if kwargs.get('project_name') and not (
            kwargs.get('project_domain_name')
            or kwargs.get('project_domain_id')
        ):
            m = (
                "You have provided a project_name. In the V3 identity API a "
                "project_name is only unique within a domain so you must "
                "also provide either a project_domain_id or "
                "project_domain_name."
            )
            raise exceptions.OptionError(m)

        return super().load_from_options(**kwargs)


class BaseFederationLoader(BaseV3Loader[V3FederationPluginT]):
    """Base Option handling for federation plugins.

    This class defines options and handling that should be common to the V3
    identity federation API. It provides the options expected by the
    :py:class:`keystoneauth1.identity.v3.FederationBaseAuth` class.
    """

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                opts.Opt(
                    'identity-provider',
                    help="Identity Provider's name",
                    required=True,
                ),
                opts.Opt(
                    'protocol',
                    help='Protocol for federated plugin',
                    required=True,
                ),
            ]
        )

        return options


class BaseGenericLoader(BaseIdentityLoader[GenericPluginT]):
    """Base Option handling for generic plugins.

    This class defines options and handling that should be common to generic
    plugins. These plugins target the OpenStack identity service however are
    designed to be independent of API version. It provides the options expected
    by the :py:class:`keystoneauth1.identity.generic.base.BaseGenericPlugin`
    class.
    """

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                opts.Opt('system-scope', help='Scope for system operations'),
                opts.Opt('domain-id', help='Domain ID to scope to'),
                opts.Opt('domain-name', help='Domain name to scope to'),
                opts.Opt(
                    'project-id',
                    help='Project ID to scope to',
                    deprecated=[opts.Opt('tenant-id')],
                ),
                opts.Opt(
                    'project-name',
                    help='Project name to scope to',
                    deprecated=[opts.Opt('tenant-name')],
                ),
                opts.Opt(
                    'project-domain-id', help='Domain ID containing project'
                ),
                opts.Opt(
                    'project-domain-name',
                    help='Domain name containing project',
                ),
                opts.Opt(
                    'trust-id', help='ID of the trust to use as a trustee use'
                ),
                opts.Opt(
                    'default-domain-id',
                    help=(
                        'Optional domain ID to use with v3 and v2 '
                        'parameters. It will be used for both the user '
                        'and project domain in v3 and ignored in '
                        'v2 authentication.'
                    ),
                ),
                opts.Opt(
                    'default-domain-name',
                    help=(
                        'Optional domain name to use with v3 API and v2 '
                        'parameters. It will be used for both the user '
                        'and project domain in v3 and ignored in '
                        'v2 authentication.'
                    ),
                ),
            ]
        )

        return options



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import fixture
from keystoneauth1.tests.unit import utils


class V2TokenTests(utils.TestCase):
    def test_unscoped(self):
        token_id = uuid.uuid4().hex
        user_id = uuid.uuid4().hex
        user_name = uuid.uuid4().hex

        token = fixture.V2Token(
            token_id=token_id, user_id=user_id, user_name=user_name
        )

        self.assertEqual(token_id, token.token_id)
        self.assertEqual(token_id, token['access']['token']['id'])
        self.assertEqual(user_id, token.user_id)
        self.assertEqual(user_id, token['access']['user']['id'])
        self.assertEqual(user_name, token.user_name)
        self.assertEqual(user_name, token['access']['user']['name'])
        self.assertIsNone(token.trust_id)

    def test_tenant_scoped(self):
        tenant_id = uuid.uuid4().hex
        tenant_name = uuid.uuid4().hex

        token = fixture.V2Token(tenant_id=tenant_id, tenant_name=tenant_name)

        self.assertEqual(tenant_id, token.tenant_id)
        self.assertEqual(tenant_id, token['access']['token']['tenant']['id'])
        self.assertEqual(tenant_name, token.tenant_name)
        tn = token['access']['token']['tenant']['name']
        self.assertEqual(tenant_name, tn)
        self.assertIsNone(token.trust_id)

    def test_trust_scoped(self):
        trust_id = uuid.uuid4().hex
        trustee_user_id = uuid.uuid4().hex

        token = fixture.V2Token(
            trust_id=trust_id, trustee_user_id=trustee_user_id
        )
        trust = token['access']['trust']

        self.assertEqual(trust_id, token.trust_id)
        self.assertEqual(trust_id, trust['id'])
        self.assertEqual(trustee_user_id, token.trustee_user_id)
        self.assertEqual(trustee_user_id, trust['trustee_user_id'])

    def test_roles(self):
        role_id1 = uuid.uuid4().hex
        role_name1 = uuid.uuid4().hex
        role_id2 = uuid.uuid4().hex
        role_name2 = uuid.uuid4().hex

        token = fixture.V2Token()
        token.add_role(id=role_id1, name=role_name1)
        token.add_role(id=role_id2, name=role_name2)

        role_names = token['access']['user']['roles']
        role_ids = token['access']['metadata']['roles']

        self.assertEqual({role_id1, role_id2}, set(role_ids))
        for r in (role_name1, role_name2):
            self.assertIn({'name': r}, role_names)

    def test_services(self):
        service_type = uuid.uuid4().hex
        service_name = uuid.uuid4().hex
        endpoint_id = uuid.uuid4().hex
        region = uuid.uuid4().hex

        public = uuid.uuid4().hex
        admin = uuid.uuid4().hex
        internal = uuid.uuid4().hex

        token = fixture.V2Token()
        svc = token.add_service(type=service_type, name=service_name)

        svc.add_endpoint(
            public=public,
            admin=admin,
            internal=internal,
            region=region,
            id=endpoint_id,
        )

        self.assertEqual(1, len(token['access']['serviceCatalog']))
        service = token['access']['serviceCatalog'][0]['endpoints'][0]

        self.assertEqual(public, service['publicURL'])
        self.assertEqual(internal, service['internalURL'])
        self.assertEqual(admin, service['adminURL'])
        self.assertEqual(region, service['region'])
        self.assertEqual(endpoint_id, service['id'])

        token.remove_service(type=service_type)
        self.assertEqual(0, len(token['access']['serviceCatalog']))

    def test_token_bind(self):
        name1 = uuid.uuid4().hex
        data1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex
        data2 = {uuid.uuid4().hex: uuid.uuid4().hex}

        token = fixture.V2Token()
        token.set_bind(name1, data1)
        token.set_bind(name2, data2)

        self.assertEqual(
            {name1: data1, name2: data2}, token['access']['token']['bind']
        )


class V3TokenTests(utils.TestCase):
    def test_unscoped(self):
        user_id = uuid.uuid4().hex
        user_name = uuid.uuid4().hex
        user_domain_id = uuid.uuid4().hex
        user_domain_name = uuid.uuid4().hex

        token = fixture.V3Token(
            user_id=user_id,
            user_name=user_name,
            user_domain_id=user_domain_id,
            user_domain_name=user_domain_name,
        )

        self.assertEqual(user_id, token.user_id)
        self.assertEqual(user_id, token['token']['user']['id'])
        self.assertEqual(user_name, token.user_name)
        self.assertEqual(user_name, token['token']['user']['name'])

        user_domain = token['token']['user']['domain']

        self.assertEqual(user_domain_id, token.user_domain_id)
        self.assertEqual(user_domain_id, user_domain['id'])
        self.assertEqual(user_domain_name, token.user_domain_name)
        self.assertEqual(user_domain_name, user_domain['name'])

    def test_project_scoped(self):
        project_id = uuid.uuid4().hex
        project_name = uuid.uuid4().hex
        project_domain_id = uuid.uuid4().hex
        project_domain_name = uuid.uuid4().hex

        token = fixture.V3Token(
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
        )

        self.assertEqual(project_id, token.project_id)
        self.assertEqual(project_id, token['token']['project']['id'])
        self.assertEqual(project_name, token.project_name)
        self.assertEqual(project_name, token['token']['project']['name'])

        self.assertIsNone(token.get('token', {}).get('is_domain'))

        project_domain = token['token']['project']['domain']

        self.assertEqual(project_domain_id, token.project_domain_id)
        self.assertEqual(project_domain_id, project_domain['id'])
        self.assertEqual(project_domain_name, token.project_domain_name)
        self.assertEqual(project_domain_name, project_domain['name'])

    def test_project_as_domain_scoped(self):
        project_id = uuid.uuid4().hex
        project_name = uuid.uuid4().hex
        project_domain_id = uuid.uuid4().hex
        project_domain_name = uuid.uuid4().hex
        project_is_domain = True

        token = fixture.V3Token(
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            project_is_domain=project_is_domain,
        )

        self.assertEqual(project_id, token.project_id)
        self.assertEqual(project_id, token['token']['project']['id'])
        self.assertEqual(project_name, token.project_name)
        self.assertEqual(project_name, token['token']['project']['name'])
        self.assertEqual(project_is_domain, token['token']['is_domain'])

        project_domain = token['token']['project']['domain']

        self.assertEqual(project_domain_id, token.project_domain_id)
        self.assertEqual(project_domain_id, project_domain['id'])
        self.assertEqual(project_domain_name, token.project_domain_name)
        self.assertEqual(project_domain_name, project_domain['name'])

    def test_domain_scoped(self):
        domain_id = uuid.uuid4().hex
        domain_name = uuid.uuid4().hex

        token = fixture.V3Token(domain_id=domain_id, domain_name=domain_name)

        self.assertEqual(domain_id, token.domain_id)
        self.assertEqual(domain_id, token['token']['domain']['id'])
        self.assertEqual(domain_name, token.domain_name)
        self.assertEqual(domain_name, token['token']['domain']['name'])

    def test_roles(self):
        role1 = {'id': uuid.uuid4().hex, 'name': uuid.uuid4().hex}
        role2 = {'id': uuid.uuid4().hex, 'name': uuid.uuid4().hex}

        token = fixture.V3Token()
        token.add_role(**role1)
        token.add_role(**role2)

        self.assertEqual(2, len(token['token']['roles']))

        self.assertIn(role1, token['token']['roles'])
        self.assertIn(role2, token['token']['roles'])

    def test_trust_scoped(self):
        trust_id = uuid.uuid4().hex
        trustee_user_id = uuid.uuid4().hex
        trustor_user_id = uuid.uuid4().hex
        impersonation = True

        token = fixture.V3Token(
            trust_id=trust_id,
            trustee_user_id=trustee_user_id,
            trustor_user_id=trustor_user_id,
            trust_impersonation=impersonation,
        )

        trust = token['token']['OS-TRUST:trust']
        self.assertEqual(trust_id, token.trust_id)
        self.assertEqual(trust_id, trust['id'])
        self.assertEqual(trustee_user_id, token.trustee_user_id)
        self.assertEqual(trustee_user_id, trust['trustee_user']['id'])
        self.assertEqual(trustor_user_id, token.trustor_user_id)
        self.assertEqual(trustor_user_id, trust['trustor_user']['id'])
        self.assertEqual(impersonation, token.trust_impersonation)
        self.assertEqual(impersonation, trust['impersonation'])

    def test_oauth_scoped(self):
        access_id = uuid.uuid4().hex
        consumer_id = uuid.uuid4().hex

        token = fixture.V3Token(
            oauth_access_token_id=access_id, oauth_consumer_id=consumer_id
        )

        oauth = token['token']['OS-OAUTH1']

        self.assertEqual(access_id, token.oauth_access_token_id)
        self.assertEqual(access_id, oauth['access_token_id'])
        self.assertEqual(consumer_id, token.oauth_consumer_id)
        self.assertEqual(consumer_id, oauth['consumer_id'])

    def test_catalog(self):
        service_type = uuid.uuid4().hex
        service_name = uuid.uuid4().hex
        service_id = uuid.uuid4().hex
        region = uuid.uuid4().hex
        endpoints = {
            'public': uuid.uuid4().hex,
            'internal': uuid.uuid4().hex,
            'admin': uuid.uuid4().hex,
        }

        token = fixture.V3Token()
        svc = token.add_service(
            type=service_type, name=service_name, id=service_id
        )
        svc.add_standard_endpoints(region=region, **endpoints)

        self.assertEqual(1, len(token['token']['catalog']))
        service = token['token']['catalog'][0]
        self.assertEqual(3, len(service['endpoints']))

        self.assertEqual(service_name, service['name'])
        self.assertEqual(service_type, service['type'])
        self.assertEqual(service_id, service['id'])

        for endpoint in service['endpoints']:
            # assert an id exists for each endpoint, remove it to make testing
            # the endpoint content below easier.
            self.assertTrue(endpoint.pop('id'))

        for interface, url in endpoints.items():
            endpoint = {
                'interface': interface,
                'url': url,
                'region': region,
                'region_id': region,
            }
            self.assertIn(endpoint, service['endpoints'])

        token.remove_service(type=service_type)
        self.assertEqual(0, len(token['token']['catalog']))

    def test_empty_default_service_providers(self):
        token = fixture.V3Token()
        self.assertIsNone(token.service_providers)

    def test_service_providers(self):
        def new_sp():
            return {
                'id': uuid.uuid4().hex,
                'sp_url': uuid.uuid4().hex,
                'auth_url': uuid.uuid4().hex,
            }

        ref_service_providers = [new_sp(), new_sp()]
        token = fixture.V3Token()
        for sp in ref_service_providers:
            token.add_service_provider(sp['id'], sp['auth_url'], sp['sp_url'])
        self.assertEqual(ref_service_providers, token.service_providers)
        self.assertEqual(
            ref_service_providers, token['token']['service_providers']
        )

    def test_token_bind(self):
        name1 = uuid.uuid4().hex
        data1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex
        data2 = {uuid.uuid4().hex: uuid.uuid4().hex}

        token = fixture.V3Token()
        token.set_bind(name1, data1)
        token.set_bind(name2, data2)

        self.assertEqual({name1: data1, name2: data2}, token['token']['bind'])

    def test_is_admin_project(self):
        token = fixture.V3Token()
        self.assertIsNone(token.is_admin_project)
        self.assertNotIn('is_admin_project', token['token'])

        token.is_admin_project = True
        self.assertIs(True, token.is_admin_project)
        self.assertIs(True, token['token']['is_admin_project'])

        token.is_admin_project = False
        self.assertIs(False, token.is_admin_project)
        self.assertIs(False, token['token']['is_admin_project'])

        del token.is_admin_project
        self.assertIsNone(token.is_admin_project)
        self.assertNotIn('is_admin_project', token['token'])

    def test_oauth2(self):
        methods = ['oauth2_credential']
        oauth2_thumbprint = uuid.uuid4().hex
        token = fixture.V3Token(
            methods=methods, oauth2_thumbprint=oauth2_thumbprint
        )
        oauth2_credential = {'x5t#S256': oauth2_thumbprint}
        self.assertEqual(methods, token.methods)
        self.assertEqual(oauth2_credential, token.oauth2_credential)
        self.assertEqual(oauth2_thumbprint, token.oauth2_thumbprint)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import uuid

from oslo_utils import timeutils

from keystoneauth1 import access
from keystoneauth1 import fixture
from keystoneauth1.tests.unit import utils


class AccessV3Test(utils.TestCase):
    def test_building_unscoped_accessinfo(self):
        token = fixture.V3Token()
        token_id = uuid.uuid4().hex

        auth_ref = access.create(body=token, auth_token=token_id)

        self.assertIn('methods', auth_ref._data['token'])
        self.assertFalse(auth_ref.has_service_catalog())
        self.assertNotIn('catalog', auth_ref._data['token'])

        self.assertEqual(token_id, auth_ref.auth_token)
        self.assertEqual(token.user_name, auth_ref.username)
        self.assertEqual(token.user_id, auth_ref.user_id)

        self.assertEqual(auth_ref.role_ids, [])
        self.assertEqual(auth_ref.role_names, [])

        self.assertIsNone(auth_ref.project_name)
        self.assertIsNone(auth_ref.project_id)

        self.assertFalse(auth_ref.domain_scoped)
        self.assertFalse(auth_ref.project_scoped)
        self.assertIsNone(auth_ref.project_is_domain)

        self.assertEqual(token.user_domain_id, auth_ref.user_domain_id)
        self.assertEqual(token.user_domain_name, auth_ref.user_domain_name)

        self.assertIsNone(auth_ref.project_domain_id)
        self.assertIsNone(auth_ref.project_domain_name)

        self.assertEqual(
            auth_ref.expires,
            timeutils.parse_isotime(token['token']['expires_at']),
        )
        self.assertEqual(
            auth_ref.issued,
            timeutils.parse_isotime(token['token']['issued_at']),
        )

        self.assertEqual(auth_ref.expires, token.expires)
        self.assertEqual(auth_ref.issued, token.issued)

        self.assertEqual(auth_ref.audit_id, token.audit_id)
        self.assertIsNone(auth_ref.audit_chain_id)
        self.assertIsNone(token.audit_chain_id)
        self.assertIsNone(auth_ref.bind)

    def test_will_expire_soon(self):
        expires = timeutils.utcnow() + datetime.timedelta(minutes=5)
        token = fixture.V3Token(expires=expires)
        auth_ref = access.create(body=token)
        self.assertFalse(auth_ref.will_expire_soon(stale_duration=120))
        self.assertTrue(auth_ref.will_expire_soon(stale_duration=301))
        self.assertFalse(auth_ref.will_expire_soon())

    def test_building_system_scoped_assessinfo(self):
        token = fixture.V3Token()
        token.set_system_scope()

        s = token.add_service(type='identity')
        s.add_standard_endpoints(public='http://url')

        token_id = uuid.uuid4().hex

        auth_ref = access.create(body=token, auth_token=token_id)

        self.assertTrue(auth_ref)
        self.assertIn('methods', auth_ref._data['token'])
        self.assertIn('catalog', auth_ref._data['token'])
        self.assertTrue(auth_ref.has_service_catalog())
        self.assertTrue(auth_ref._data['token']['catalog'])

        self.assertEqual(token_id, auth_ref.auth_token)
        self.assertEqual(token.user_name, auth_ref.username)
        self.assertEqual(token.user_id, auth_ref.user_id)

        self.assertEqual(token.role_ids, auth_ref.role_ids)
        self.assertEqual(token.role_names, auth_ref.role_names)

        self.assertEqual(token.domain_name, auth_ref.domain_name)
        self.assertEqual(token.domain_id, auth_ref.domain_id)

        self.assertEqual(token.user_domain_id, auth_ref.user_domain_id)
        self.assertEqual(token.user_domain_name, auth_ref.user_domain_name)

        self.assertIsNone(auth_ref.project_name)
        self.assertIsNone(auth_ref.project_id)

        self.assertIsNone(auth_ref.project_domain_id)
        self.assertIsNone(auth_ref.project_domain_name)

        self.assertIsNone(auth_ref.domain_name)
        self.assertIsNone(auth_ref.domain_id)

        self.assertEqual(token.system, auth_ref.system)

        self.assertTrue(auth_ref.system_scoped)
        self.assertFalse(auth_ref.domain_scoped)
        self.assertFalse(auth_ref.project_scoped)

        self.assertEqual(token.audit_id, auth_ref.audit_id)
        self.assertEqual(token.audit_chain_id, auth_ref.audit_chain_id)

    def test_building_domain_scoped_accessinfo(self):
        token = fixture.V3Token()
        token.set_domain_scope()

        s = token.add_service(type='identity')
        s.add_standard_endpoints(public='http://url')

        token_id = uuid.uuid4().hex

        auth_ref = access.create(body=token, auth_token=token_id)

        self.assertTrue(auth_ref)
        self.assertIn('methods', auth_ref._data['token'])
        self.assertIn('catalog', auth_ref._data['token'])
        self.assertTrue(auth_ref.has_service_catalog())
        self.assertTrue(auth_ref._data['token']['catalog'])

        self.assertEqual(token_id, auth_ref.auth_token)
        self.assertEqual(token.user_name, auth_ref.username)
        self.assertEqual(token.user_id, auth_ref.user_id)

        self.assertEqual(token.role_ids, auth_ref.role_ids)
        self.assertEqual(token.role_names, auth_ref.role_names)

        self.assertEqual(token.domain_name, auth_ref.domain_name)
        self.assertEqual(token.domain_id, auth_ref.domain_id)

        self.assertIsNone(auth_ref.project_name)
        self.assertIsNone(auth_ref.project_id)

        self.assertEqual(token.user_domain_id, auth_ref.user_domain_id)
        self.assertEqual(token.user_domain_name, auth_ref.user_domain_name)

        self.assertIsNone(auth_ref.project_domain_id)
        self.assertIsNone(auth_ref.project_domain_name)

        self.assertTrue(auth_ref.domain_scoped)
        self.assertFalse(auth_ref.project_scoped)
        self.assertIsNone(auth_ref.project_is_domain)

        self.assertEqual(token.audit_id, auth_ref.audit_id)
        self.assertEqual(token.audit_chain_id, auth_ref.audit_chain_id)

    def test_building_project_scoped_accessinfo(self):
        token = fixture.V3Token()
        token.set_project_scope()

        s = token.add_service(type='identity')
        s.add_standard_endpoints(public='http://url')

        token_id = uuid.uuid4().hex

        auth_ref = access.create(body=token, auth_token=token_id)

        self.assertIn('methods', auth_ref._data['token'])
        self.assertIn('catalog', auth_ref._data['token'])
        self.assertTrue(auth_ref.has_service_catalog())
        self.assertTrue(auth_ref._data['token']['catalog'])

        self.assertEqual(token_id, auth_ref.auth_token)
        self.assertEqual(token.user_name, auth_ref.username)
        self.assertEqual(token.user_id, auth_ref.user_id)

        self.assertEqual(token.role_ids, auth_ref.role_ids)
        self.assertEqual(token.role_names, auth_ref.role_names)

        self.assertIsNone(auth_ref.domain_name)
        self.assertIsNone(auth_ref.domain_id)

        self.assertEqual(token.project_name, auth_ref.project_name)
        self.assertEqual(token.project_id, auth_ref.project_id)

        self.assertEqual(auth_ref.tenant_name, auth_ref.project_name)
        self.assertEqual(auth_ref.tenant_id, auth_ref.project_id)

        self.assertEqual(token.project_domain_id, auth_ref.project_domain_id)
        self.assertEqual(
            token.project_domain_name, auth_ref.project_domain_name
        )

        self.assertEqual(token.user_domain_id, auth_ref.user_domain_id)
        self.assertEqual(token.user_domain_name, auth_ref.user_domain_name)

        self.assertFalse(auth_ref.domain_scoped)
        self.assertTrue(auth_ref.project_scoped)
        self.assertIsNone(auth_ref.project_is_domain)

        self.assertEqual(token.audit_id, auth_ref.audit_id)
        self.assertEqual(token.audit_chain_id, auth_ref.audit_chain_id)

    def test_building_project_as_domain_scoped_accessinfo(self):
        token = fixture.V3Token()
        token.set_project_scope(is_domain=True)

        service = token.add_service(type='identity')
        service.add_standard_endpoints(public='http://url')

        token_id = uuid.uuid4().hex

        auth_ref = access.create(body=token, auth_token=token_id)

        self.assertIn('methods', auth_ref._data['token'])
        self.assertIn('catalog', auth_ref._data['token'])
        self.assertTrue(auth_ref.has_service_catalog())
        self.assertTrue(auth_ref._data['token']['catalog'])

        self.assertEqual(token_id, auth_ref.auth_token)
        self.assertEqual(token.user_name, auth_ref.username)
        self.assertEqual(token.user_id, auth_ref.user_id)

        self.assertEqual(token.role_ids, auth_ref.role_ids)
        self.assertEqual(token.role_names, auth_ref.role_names)

        self.assertIsNone(auth_ref.domain_name)
        self.assertIsNone(auth_ref.domain_id)

        self.assertEqual(token.project_name, auth_ref.project_name)
        self.assertEqual(token.project_id, auth_ref.project_id)

        self.assertEqual(auth_ref.tenant_name, auth_ref.project_name)
        self.assertEqual(auth_ref.tenant_id, auth_ref.project_id)

        self.assertEqual(token.project_domain_id, auth_ref.project_domain_id)
        self.assertEqual(
            token.project_domain_name, auth_ref.project_domain_name
        )

        self.assertEqual(token.user_domain_id, auth_ref.user_domain_id)
        self.assertEqual(token.user_domain_name, auth_ref.user_domain_name)

        self.assertFalse(auth_ref.domain_scoped)
        self.assertTrue(auth_ref.project_scoped)
        self.assertTrue(auth_ref.project_is_domain)

        self.assertEqual(token.audit_id, auth_ref.audit_id)
        self.assertEqual(token.audit_chain_id, auth_ref.audit_chain_id)

    def test_oauth_access(self):
        consumer_id = uuid.uuid4().hex
        access_token_id = uuid.uuid4().hex

        token = fixture.V3Token()
        token.set_project_scope()
        token.set_oauth(
            access_token_id=access_token_id, consumer_id=consumer_id
        )

        auth_ref = access.create(body=token)

        self.assertEqual(consumer_id, auth_ref.oauth_consumer_id)
        self.assertEqual(access_token_id, auth_ref.oauth_access_token_id)

        self.assertEqual(
            consumer_id, auth_ref._data['token']['OS-OAUTH1']['consumer_id']
        )
        self.assertEqual(
            access_token_id,
            auth_ref._data['token']['OS-OAUTH1']['access_token_id'],
        )

    def test_federated_property_standard_token(self):
        """Check if is_federated property returns expected value."""
        token = fixture.V3Token()
        token.set_project_scope()
        auth_ref = access.create(body=token)
        self.assertFalse(auth_ref.is_federated)

    def test_binding(self):
        token = fixture.V3Token()
        principal = uuid.uuid4().hex
        token.set_bind('kerberos', principal)

        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV3)

        self.assertEqual({'kerberos': principal}, auth_ref.bind)

    def test_is_admin_project_unset(self):
        token = fixture.V3Token()
        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV3)
        self.assertIs(True, auth_ref.is_admin_project)

    def test_is_admin_project_true(self):
        token = fixture.V3Token(is_admin_project=True)
        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV3)
        self.assertIs(True, auth_ref.is_admin_project)

    def test_is_admin_project_false(self):
        token = fixture.V3Token(is_admin_project=False)
        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV3)
        self.assertIs(False, auth_ref.is_admin_project)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# flake8: noqa: F405

from keystoneauth1.loading import adapter
from keystoneauth1.loading.base import *  # noqa
from keystoneauth1.loading import cli
from keystoneauth1.loading import conf
from keystoneauth1.loading.identity import *  # noqa
from keystoneauth1.loading.opts import *  # noqa
from keystoneauth1.loading import session


register_auth_argparse_arguments = cli.register_argparse_arguments
load_auth_from_argparse_arguments = cli.load_from_argparse_arguments

get_auth_common_conf_options = conf.get_common_conf_options
get_auth_plugin_conf_options = conf.get_plugin_conf_options
register_auth_conf_options = conf.register_conf_options
load_auth_from_conf_options = conf.load_from_conf_options

register_session_argparse_arguments = session.register_argparse_arguments
load_session_from_argparse_arguments = session.load_from_argparse_arguments
register_session_conf_options = session.register_conf_options
load_session_from_conf_options = session.load_from_conf_options
get_session_conf_options = session.get_conf_options

register_adapter_argparse_arguments = adapter.register_argparse_arguments
register_service_adapter_argparse_arguments = (
    adapter.register_service_argparse_arguments
)
register_adapter_conf_options = adapter.register_conf_options
load_adapter_from_conf_options = adapter.load_from_conf_options
get_adapter_conf_options = adapter.get_conf_options


__all__ = (
    # loading.base
    'BaseLoader',
    'get_available_plugin_names',
    'get_available_plugin_loaders',
    'get_plugin_loader',
    'PLUGIN_NAMESPACE',
    # loading.identity
    'BaseIdentityLoader',
    'BaseV2Loader',
    'BaseV3Loader',
    'BaseFederationLoader',
    'BaseGenericLoader',
    # auth cli
    'register_auth_argparse_arguments',
    'load_auth_from_argparse_arguments',
    # auth conf
    'get_auth_common_conf_options',
    'get_auth_plugin_conf_options',
    'register_auth_conf_options',
    'load_auth_from_conf_options',
    # session
    'register_session_argparse_arguments',
    'load_session_from_argparse_arguments',
    'register_session_conf_options',
    'load_session_from_conf_options',
    'get_session_conf_options',
    # adapter
    'register_adapter_argparse_arguments',
    'register_service_adapter_argparse_arguments',
    'register_adapter_conf_options',
    'load_adapter_from_conf_options',
    'get_adapter_conf_options',
    # loading.opts
    'Opt',
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""Kerberos authentication plugins.

.. warning::
    This module requires installation of an extra package (`requests_kerberos`)
    not installed by default. Without the extra package an import error will
    occur. The extra package can be installed using::

      $ pip install keystoneauth1[kerberos]
"""

import typing as ty

try:
    # explicitly re-export symbol
    # https://mypy.readthedocs.io/en/stable/command_line.html#cmdoption-mypy-no-implicit-reexport
    import requests_kerberos as requests_kerberos
except ImportError:
    requests_kerberos = None

from keystoneauth1 import access
from keystoneauth1.identity import v3
from keystoneauth1.identity.v3 import federation
from keystoneauth1 import session as ks_session


# TODO(stephenfin): This should return an enum
def _mutual_auth(value: str | None) -> str:
    default = ty.cast(str, requests_kerberos.OPTIONAL)
    if value is None:
        return default
    return {
        'required': ty.cast(str, requests_kerberos.REQUIRED),
        'optional': ty.cast(str, requests_kerberos.OPTIONAL),
        'disabled': ty.cast(str, requests_kerberos.DISABLED),
    }.get(value.lower(), default)


def _requests_auth(
    mutual_authentication: str | None,
) -> 'requests_kerberos.HTTPKerberosAuth':
    return requests_kerberos.HTTPKerberosAuth(
        mutual_authentication=_mutual_auth(mutual_authentication)
    )


def _dependency_check() -> None:
    if requests_kerberos is None:
        raise ImportError("""
Using the kerberos authentication plugin requires installation of additional
packages. These can be installed with::

    $ pip install keystoneauth1[kerberos]
""")


class KerberosMethod(v3.AuthMethod):
    mutual_auth: str | None

    def __init__(self, *, mutual_auth: str | None = None) -> None:
        self.mutual_auth = mutual_auth

        _dependency_check()

    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: v3.Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> tuple[None, None] | tuple[str, ty.Mapping[str, object]]:
        # NOTE(jamielennox): request_kwargs is passed as a kwarg however it is
        # required and always present when called from keystoneclient.
        request_kwargs['requests_auth'] = _requests_auth(self.mutual_auth)
        return 'kerberos', {}


class Kerberos(v3.Auth):
    _auth_method_class = KerberosMethod

    def __init__(
        self,
        auth_url: str,
        mutual_auth: str | None = None,
        *,
        unscoped: bool = False,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ) -> None:
        method = self._auth_method_class(mutual_auth=mutual_auth)
        super().__init__(
            auth_url,
            [method],
            unscoped=unscoped,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )


class MappedKerberos(federation.FederationBaseAuth):
    """Authenticate using Kerberos via the keystone federation mechanisms.

    This uses the OS-FEDERATION extension to gain an unscoped token and then
    use the standard keystone auth process to scope that to any given project.
    """

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        mutual_auth: str | None = None,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ) -> None:
        _dependency_check()
        self.mutual_auth = mutual_auth
        super().__init__(
            auth_url,
            identity_provider,
            protocol,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )

    def get_unscoped_auth_ref(
        self, session: ks_session.Session
    ) -> access.AccessInfoV3:
        resp = session.get(
            self.federated_token_url,
            requests_auth=_requests_auth(self.mutual_auth),
            authenticated=False,
        )

        access_info = access.create(body=resp.json(), resp=resp)
        # narrow type
        assert isinstance(access_info, access.AccessInfoV3)  # nosec B101
        return access_info



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest import mock

from keystoneauth1.loading import opts
from keystoneauth1.tests.unit.loading import utils


class OptTests(utils.TestCase):
    def test_argparse_args(self):
        opt = opts.Opt('auth-url')
        self.assertEqual(['--os-auth-url'], opt.argparse_args)

    def test_argparse_args_with_deprecations(self):
        opt = opts.Opt('username', deprecated=[opts.Opt('user-name')])
        self.assertEqual(
            ['--os-username', '--os-user-name'], opt.argparse_args
        )

    def test_argparse_envvars(self):
        opt = opts.Opt('auth-url')
        self.assertEqual(['OS_AUTH_URL'], opt.argparse_envvars)

    def test_argparse_envvars_with_deprecations(self):
        opt = opts.Opt('username', deprecated=[opts.Opt('user-name')])
        self.assertEqual(['OS_USERNAME', 'OS_USER_NAME'], opt.argparse_envvars)

    def test_argparse_default(self):
        opt = opts.Opt('auth-url')

        with mock.patch.dict(
            os.environ, {'OS_AUTH_URL': 'http://1.2.3.4/identity'}, clear=True
        ):
            self.assertEqual('http://1.2.3.4/identity', opt.argparse_default)

    def test_argparse_default_with_deprecations(self):
        opt = opts.Opt('username', deprecated=[opts.Opt('user-name')])

        with mock.patch.dict(
            os.environ, {'OS_USER_NAME': 'superuser'}, clear=True
        ):
            self.assertEqual('superuser', opt.argparse_default)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A fixture to wrap the session constructor for use with Betamax."""

import collections.abc
from functools import partial
import typing as ty
from unittest import mock

import betamax
from betamax import cassette
from betamax import serializers
import fixtures
import requests

from keystoneauth1.fixture import hooks
from keystoneauth1.fixture import serializer as yaml_serializer
from keystoneauth1 import session


PreRecordHookT: ty.TypeAlias = collections.abc.Callable[
    [cassette.Interaction, cassette.Cassette], None
]


class BetamaxFixture(fixtures.Fixture):
    recorder: betamax.Betamax

    def __init__(
        self,
        cassette_name: str,
        cassette_library_dir: str | None = None,
        serializer: type[serializers.BaseSerializer] | None = None,
        record: bool | str = False,
        pre_record_hook: PreRecordHookT = hooks.pre_record_hook,
        serializer_name: str | None = None,
        request_matchers: list[str] | None = None,
    ) -> None:
        """Configure Betamax for the test suite.

        :param str cassette_name:
            This is simply the name of the cassette without any file extension
            or containing directory. For example, to generate
            ``keystoneauth1/tests/unit/data/example.yaml``, one would pass
            only ``example``.
        :param str cassette_library_dir:
            This is the directory that will contain all cassette files. In
            ``keystoneauth1/tests/unit/data/example.yaml`` you would pass
            ``keystoneauth1/tests/unit/data/``.
        :param serializer:
            A class that implements the Serializer API in Betamax. See also:
            https://betamax.readthedocs.io/en/latest/serializers.html
        :param record:
            The Betamax record mode to use. If ``False`` (the default), then
            Betamax will not record anything. For more information about
            record modes, see:
            https://betamax.readthedocs.io/en/latest/record_modes.html
        :param callable pre_record_hook:
            Function or callable to use to perform some handling of the
            request or response data prior to saving it to disk.
        :param str serializer_name:
            The name of a serializer already registered with Betamax to use
            to handle cassettes. For example, if you want to use the default
            Betamax serializer, you would pass ``'json'`` to this parameter.
        :param list request_matchers:
            The list of request matcher names to use with Betamax. Betamax's
            default list is used if none are specified. See also:
            https://betamax.readthedocs.io/en/latest/matchers.html
        """
        self.cassette_library_dir = cassette_library_dir
        self.record = record
        self.cassette_name = cassette_name
        if not (serializer or serializer_name):
            serializer = yaml_serializer.YamlJsonSerializer
            serializer_name = serializer.name
        if serializer:
            betamax.Betamax.register_serializer(serializer)
        self.serializer = serializer
        self._serializer_name = serializer_name
        self.pre_record_hook = pre_record_hook
        self.use_cassette_kwargs = {}
        if request_matchers is not None:
            self.use_cassette_kwargs['match_requests_on'] = request_matchers

    @property
    def serializer_name(self) -> str | None:
        """Determine the name of the selected serializer.

        If a class was specified, use the name attribute to generate this,
        otherwise, use the serializer_name parameter from ``__init__``.

        :returns:
            Name of the serializer
        :rtype:
            str
        """
        if self.serializer:
            # betamax is not typed yet
            return self.serializer.name  # type: ignore
        return self._serializer_name

    def setUp(self) -> None:
        super().setUp()
        self.mockpatch = mock.patch.object(
            session,
            '_construct_session',
            partial(_construct_session_with_betamax, self),
        )
        self.mockpatch.start()
        # Unpatch during cleanup
        self.addCleanup(self.mockpatch.stop)


def _construct_session_with_betamax(
    fixture: BetamaxFixture, session_obj: requests.Session | None = None
) -> requests.Session:
    # NOTE(morganfainberg): This function should contain the logic of
    # keystoneauth1.session._construct_session as it replaces the
    # _construct_session function to apply betamax magic to the requests
    # session object.
    if not session_obj:
        session_obj = requests.Session()
        # Use TCPKeepAliveAdapter to fix bug 1323862
        for scheme in list(session_obj.adapters.keys()):
            session_obj.mount(scheme, session.TCPKeepAliveAdapter())

    with betamax.Betamax.configure() as config:
        config.before_record(callback=fixture.pre_record_hook)

    fixture.recorder = betamax.Betamax(
        session_obj, cassette_library_dir=fixture.cassette_library_dir
    )

    record = 'none'
    serializer = None

    if fixture.record in ['once', 'all', 'new_episodes']:
        # NOTE(stephenfin): idk why we need to do this, but we do
        assert isinstance(fixture.record, str)  # narrow type
        record = fixture.record

    serializer = fixture.serializer_name

    fixture.recorder.use_cassette(
        fixture.cassette_name,
        serialize_with=serializer,
        record=record,
        **fixture.use_cassette_kwargs,
    )

    fixture.recorder.start()
    fixture.addCleanup(fixture.recorder.stop)
    return session_obj



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import string

DIR = os.path.dirname(os.path.abspath(__file__))


def template(f, **kwargs):
    with open(os.path.join(DIR, 'templates', f)) as f:
        return string.Template(f.read()).substitute(**kwargs)


def soap_response(**kwargs):
    kwargs.setdefault('provider', 'https://idp.testshib.org/idp/shibboleth')
    kwargs.setdefault(
        'consumer', 'https://openstack4.local/Shibboleth.sso/SAML2/ECP'
    )
    kwargs.setdefault('issuer', 'https://openstack4.local/shibboleth')
    return template('soap_response.xml', **kwargs).encode('utf-8')


def saml_assertion(**kwargs):
    kwargs.setdefault('issuer', 'https://idp.testshib.org/idp/shibboleth')
    kwargs.setdefault(
        'destination', 'https://openstack4.local/Shibboleth.sso/SAML2/ECP'
    )
    return template('saml_assertion.xml', **kwargs).encode('utf-8')


def authn_request(**kwargs):
    kwargs.setdefault(
        'issuer', 'https://openstack4.local/Shibboleth.sso/SAML2/ECP'
    )
    return template('authn_request.xml', **kwargs).encode('utf-8')


SP_SOAP_RESPONSE = soap_response()
SAML2_ASSERTION = saml_assertion()
AUTHN_REQUEST = authn_request()

UNSCOPED_TOKEN_HEADER = 'UNSCOPED_TOKEN'

UNSCOPED_TOKEN = {
    "token": {
        "issued_at": "2014-06-09T09:48:59.643406Z",
        "extras": {},
        "methods": ["saml2"],
        "expires_at": "2014-06-09T10:48:59.643375Z",
        "user": {
            "OS-FEDERATION": {
                "identity_provider": {"id": "testshib"},
                "protocol": {"id": "saml2"},
                "groups": [{"id": "1764fa5cf69a49a4918131de5ce4af9a"}],
            },
            "id": "testhib%20user",
            "name": "testhib user",
        },
    }
}

PROJECTS = {
    "projects": [
        {
            "domain_id": "37ef61",
            "enabled": 'true',
            "id": "12d706",
            "links": {"self": "http://identity:35357/v3/projects/12d706"},
            "name": "a project name",
        },
        {
            "domain_id": "37ef61",
            "enabled": 'true',
            "id": "9ca0eb",
            "links": {"self": "http://identity:35357/v3/projects/9ca0eb"},
            "name": "another project",
        },
    ],
    "links": {
        "self": "http://identity:35357/v3/OS-FEDERATION/projects",
        "previous": 'null',
        "next": 'null',
    },
}

DOMAINS = {
    "domains": [
        {
            "description": "desc of domain",
            "enabled": 'true',
            "id": "37ef61",
            "links": {"self": "http://identity:35357/v3/domains/37ef61"},
            "name": "my domain",
        }
    ],
    "links": {
        "self": "http://identity:35357/v3/OS-FEDERATION/domains",
        "previous": 'null',
        "next": 'null',
    },
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import typing as ty

from keystoneauth1.loading import base
from keystoneauth1 import session

if ty.TYPE_CHECKING:
    from oslo_config import cfg

    from keystoneauth1.loading import opts


__all__ = (
    'register_argparse_arguments',
    'load_from_argparse_arguments',
    'register_conf_options',
    'load_from_conf_options',
    'get_conf_options',
)


def _positive_non_zero_float(argument_value: str | None) -> float | None:
    if argument_value is None:
        return None
    try:
        value = float(argument_value)
    except ValueError:
        msg = f"{argument_value} must be a float"
        raise argparse.ArgumentTypeError(msg)
    if value <= 0:
        msg = f"{argument_value} must be greater than 0"
        raise argparse.ArgumentTypeError(msg)
    return value


class Session(base._BaseLoader[session.Session]):
    @property
    def plugin_class(self) -> type[session.Session]:
        return session.Session

    def get_options(self) -> list['opts.Opt']:
        return []

    def load_from_options(
        self,
        insecure: bool = False,
        verify: bool | str | None = None,
        cacert: str | None = None,
        cert: str | None = None,
        key: str | None = None,
        **kwargs: ty.Any,
    ) -> session.Session:
        """Create a session with individual certificate parameters.

        Some parameters used to create a session don't lend themselves to be
        loaded from config/CLI etc. Create a session by converting those
        parameters into session __init__ parameters.
        """
        if verify is None:
            if insecure:
                verify = False
            else:
                verify = cacert or True

        cert_: str | None | tuple[str, str] = cert
        if cert and key:
            # passing cert and key together is deprecated in favour of the
            # requests lib form of having the cert and key as a tuple
            cert_ = (cert, key)

        return super().load_from_options(verify=verify, cert=cert_, **kwargs)

    def register_argparse_arguments(
        self, parser: argparse.ArgumentParser
    ) -> None:
        session_group = parser.add_argument_group(
            'API Connection Options',
            'Options controlling the HTTP API Connections',
        )

        session_group.add_argument(
            '--insecure',
            default=False,
            action='store_true',
            help='Explicitly allow client to perform '
            '"insecure" TLS (https) requests. The '
            'server\'s certificate will not be verified '
            'against any certificate authorities. This '
            'option should be used with caution.',
        )

        session_group.add_argument(
            '--os-cacert',
            metavar='<ca-certificate>',
            default=os.environ.get('OS_CACERT'),
            help='Specify a CA bundle file to use in '
            'verifying a TLS (https) server certificate. '
            'Defaults to env[OS_CACERT].',
        )

        session_group.add_argument(
            '--os-cert',
            metavar='<certificate>',
            default=os.environ.get('OS_CERT'),
            help='The location for the keystore (PEM formatted) '
            'containing the public key of this client. '
            'Defaults to env[OS_CERT].',
        )

        session_group.add_argument(
            '--os-key',
            metavar='<key>',
            default=os.environ.get('OS_KEY'),
            help='The location for the keystore (PEM formatted) '
            'containing the private key of this client. '
            'Defaults to env[OS_KEY].',
        )

        session_group.add_argument(
            '--timeout',
            default=600,
            type=_positive_non_zero_float,
            metavar='<seconds>',
            help='Set request timeout (in seconds).',
        )

        session_group.add_argument(
            '--collect-timing',
            default=False,
            action='store_true',
            help='Collect per-API call timing information.',
        )

    def load_from_argparse_arguments(
        self, namespace: argparse.Namespace, **kwargs: ty.Any
    ) -> session.Session:
        kwargs.setdefault('insecure', namespace.insecure)
        kwargs.setdefault('cacert', namespace.os_cacert)
        kwargs.setdefault('cert', namespace.os_cert)
        kwargs.setdefault('key', namespace.os_key)
        kwargs.setdefault('timeout', namespace.timeout)
        kwargs.setdefault('collect_timing', namespace.collect_timing)

        return self.load_from_options(**kwargs)

    def get_conf_options(
        self,
        deprecated_opts: dict[str, list['cfg.DeprecatedOpt']] | None = None,
    ) -> list['cfg.Opt']:
        """Get oslo_config options that are needed for a :py:class:`.Session`.

        These may be useful without being registered for config file generation
        or to manipulate the options before registering them yourself.

        The options that are set are:
            :cafile: The certificate authority filename.
            :certfile: The client certificate file to present.
            :keyfile: The key for the client certificate.
            :insecure: Whether to ignore SSL verification.
            :timeout: The max time to wait for HTTP connections.
            :collect-timing: Whether to collect API timing information.
            :split-loggers: Whether to log requests to multiple loggers.

        :param dict deprecated_opts: Deprecated options that should be included
             in the definition of new options. This should be a dict from the
             name of the new option to a list of oslo.DeprecatedOpts that
             correspond to the new option. (optional)

             For example, to support the ``ca_file`` option pointing to the new
             ``cafile`` option name::

                 old_opt = oslo_cfg.DeprecatedOpt('ca_file', 'old_group')
                 deprecated_opts = {'cafile': [old_opt]}

        :returns: A list of oslo_config options.
        """
        from oslo_config import cfg

        if deprecated_opts is None:
            deprecated_opts = {}

        return [
            cfg.StrOpt(
                'cafile',
                deprecated_opts=deprecated_opts.get('cafile'),
                help='PEM encoded Certificate Authority to use '
                'when verifying HTTPs connections.',
            ),
            cfg.StrOpt(
                'certfile',
                deprecated_opts=deprecated_opts.get('certfile'),
                help='PEM encoded client certificate cert file',
            ),
            cfg.StrOpt(
                'keyfile',
                deprecated_opts=deprecated_opts.get('keyfile'),
                help='PEM encoded client certificate key file',
            ),
            cfg.BoolOpt(
                'insecure',
                default=False,
                deprecated_opts=deprecated_opts.get('insecure'),
                help='Verify HTTPS connections.',
            ),
            cfg.FloatOpt(
                'timeout',
                deprecated_opts=deprecated_opts.get('timeout'),
                help='Timeout value for http requests',
            ),
            cfg.BoolOpt(
                'collect-timing',
                deprecated_opts=deprecated_opts.get('collect-timing'),
                default=False,
                help='Collect per-API call timing information.',
            ),
            cfg.BoolOpt(
                'split-loggers',
                deprecated_opts=deprecated_opts.get('split-loggers'),
                default=False,
                help='Log requests to multiple loggers.',
            ),
        ]

    def register_conf_options(
        self,
        conf: 'cfg.ConfigOpts',
        group: str,
        deprecated_opts: dict[str, list['cfg.DeprecatedOpt']] | None = None,
    ) -> list['cfg.Opt']:
        """Register the oslo_config options that are needed for a session.

        The options that are set are:
            :cafile: The certificate authority filename.
            :certfile: The client certificate file to present.
            :keyfile: The key for the client certificate.
            :insecure: Whether to ignore SSL verification.
            :timeout: The max time to wait for HTTP connections.
            :collect-timing: Whether to collect API timing information.
            :split-loggers: Whether to log requests to multiple loggers.

        :param oslo_config.Cfg conf: config object to register with.
        :param string group: The ini group to register options in.
        :param dict deprecated_opts: Deprecated options that should be included
             in the definition of new options. This should be a dict from the
             name of the new option to a list of oslo.DeprecatedOpts that
             correspond to the new option. (optional)

             For example, to support the ``ca_file`` option pointing to the new
             ``cafile`` option name::

                 old_opt = oslo_cfg.DeprecatedOpt('ca_file', 'old_group')
                 deprecated_opts = {'cafile': [old_opt]}

        :returns: The list of options that was registered.
        """
        from oslo_config import cfg

        opts = self.get_conf_options(deprecated_opts=deprecated_opts)
        conf.register_group(cfg.OptGroup(group))
        conf.register_opts(opts, group=group)
        return opts

    def load_from_conf_options(
        self, conf: 'cfg.ConfigOpts', group: str, **kwargs: ty.Any
    ) -> session.Session:
        """Create a session object from an oslo_config object.

        The options must have been previously registered with
        register_conf_options.

        :param oslo_config.Cfg conf: config object to register with.
        :param string group: The ini group to register options in.
        :param dict kwargs: Additional parameters to pass to session
                            construction.
        :returns: A new session object.
        :rtype: :py:class:`.Session`
        """
        c = conf[group]

        kwargs.setdefault('insecure', c.insecure)
        kwargs.setdefault('cacert', c.cafile)
        kwargs.setdefault('cert', c.certfile)
        kwargs.setdefault('key', c.keyfile)
        kwargs.setdefault('timeout', c.timeout)
        kwargs.setdefault('collect_timing', c.collect_timing)
        kwargs.setdefault('split_loggers', c.split_loggers)

        return self.load_from_options(**kwargs)


def register_argparse_arguments(parser: argparse.ArgumentParser) -> None:
    Session().register_argparse_arguments(parser)


def load_from_argparse_arguments(
    namespace: argparse.Namespace, **kwargs: ty.Any
) -> session.Session:
    return Session().load_from_argparse_arguments(namespace, **kwargs)


def register_conf_options(
    conf: 'cfg.ConfigOpts',
    group: str,
    deprecated_opts: dict[str, list['cfg.DeprecatedOpt']] | None = None,
) -> list['cfg.Opt']:
    return Session().register_conf_options(
        conf, group, deprecated_opts=deprecated_opts
    )


def load_from_conf_options(
    conf: 'cfg.ConfigOpts', group: str, **kwargs: ty.Any
) -> session.Session:
    return Session().load_from_conf_options(conf, group, **kwargs)


def get_conf_options(
    deprecated_opts: dict[str, list['cfg.DeprecatedOpt']] | None = None,
) -> list['cfg.Opt']:
    return Session().get_conf_options(deprecated_opts=deprecated_opts)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.loading._plugins import noauth as loader
from keystoneauth1 import noauth
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils


class NoAuthTest(utils.TestCase):
    NOAUTH_TOKEN = 'notused'
    TEST_URL = 'http://server/prefix'

    def test_basic_case(self):
        self.requests_mock.get(self.TEST_URL, text='body')

        a = noauth.NoAuth()
        s = session.Session(auth=a)

        data = s.get(self.TEST_URL, authenticated=True)

        self.assertEqual(data.text, 'body')
        self.assertRequestHeaderEqual('X-Auth-Token', self.NOAUTH_TOKEN)
        self.assertIsNone(a.get_endpoint(s))

    def test_noauth_options(self):
        opts = loader.NoAuth().get_options()
        self.assertEqual(['endpoint'], [o.name for o in opts])

    def test_get_endpoint(self):
        a = noauth.NoAuth(endpoint=self.TEST_URL)
        s = session.Session(auth=a)
        self.assertEqual(self.TEST_URL, a.get_endpoint(s))

    def test_get_endpoint_with_override(self):
        a = noauth.NoAuth(endpoint=self.TEST_URL)
        s = session.Session(auth=a)
        self.assertEqual('foo', a.get_endpoint(s, endpoint_override='foo'))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import typing as ty

from keystoneauth1 import plugin

if ty.TYPE_CHECKING:
    from keystoneauth1 import session as ks_session

AUTH_HEADER_NAME = 'Authorization'


class HTTPBasicAuth(plugin.FixedEndpointPlugin):
    """A provider that will always use HTTP Basic authentication.

    This is useful to unify session/adapter loading for services
    that might be deployed in standalone mode.
    """

    def __init__(
        self,
        endpoint: str | None = None,
        username: str | None = None,
        password: str | None = None,
    ):
        super().__init__(endpoint)
        self.username = username
        self.password = password

    def get_token(self, session: 'ks_session.Session') -> str | None:
        if self.username is None or self.password is None:
            return None
        token = bytes(f'{self.username}:{self.password}', encoding='utf-8')
        encoded = base64.b64encode(token)
        return str(encoded, encoding='utf-8')

    def get_headers(
        self, session: 'ks_session.Session'
    ) -> dict[str, str] | None:
        token = self.get_token(session)
        if not token:
            return None
        auth = f'Basic {token}'
        return {AUTH_HEADER_NAME: auth}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.extras import kerberos
from keystoneauth1 import session
from keystoneauth1.tests.unit.extras.kerberos import base


class TestKerberosAuth(base.TestCase):
    def setUp(self):
        if kerberos.requests_kerberos is None:
            self.skipTest("Kerberos support isn't available.")

        super().setUp()

    def test_authenticate_with_kerberos_domain_scoped(self):
        token_id, token_body = self.kerberos_mock.mock_auth_success()

        a = kerberos.Kerberos(self.TEST_ROOT_URL + 'v3')
        s = session.Session(a)
        token = a.get_token(s)

        self.assertRequestBody()
        self.assertEqual(
            self.kerberos_mock.challenge_header,
            self.requests_mock.last_request.headers['Authorization'],
        )
        self.assertEqual(token_id, a.auth_ref.auth_token)
        self.assertEqual(token_id, token)

    def test_authenticate_with_kerberos_mutual_authentication_required(self):
        token_id, token_body = self.kerberos_mock.mock_auth_success()

        a = kerberos.Kerberos(
            self.TEST_ROOT_URL + 'v3', mutual_auth='required'
        )
        s = session.Session(a)
        token = a.get_token(s)

        self.assertRequestBody()
        self.assertEqual(
            self.kerberos_mock.challenge_header,
            self.requests_mock.last_request.headers['Authorization'],
        )
        self.assertEqual(token_id, a.auth_ref.auth_token)
        self.assertEqual(token_id, token)
        self.assertEqual(self.kerberos_mock.called_auth_server, True)

    def test_authenticate_with_kerberos_mutual_authentication_disabled(self):
        token_id, token_body = self.kerberos_mock.mock_auth_success()

        a = kerberos.Kerberos(
            self.TEST_ROOT_URL + 'v3', mutual_auth='disabled'
        )
        s = session.Session(a)
        token = a.get_token(s)

        self.assertRequestBody()
        self.assertEqual(
            self.kerberos_mock.challenge_header,
            self.requests_mock.last_request.headers['Authorization'],
        )
        self.assertEqual(token_id, a.auth_ref.auth_token)
        self.assertEqual(token_id, token)
        self.assertEqual(self.kerberos_mock.called_auth_server, False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import uuid

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1 import fixture
from keystoneauth1 import identity
from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneauth1.tests.unit import k2k_fixtures
from keystoneauth1.tests.unit import utils


class TesterFederationPlugin(v3.FederationBaseAuth):
    def get_unscoped_auth_ref(self, sess):
        # This would go and talk to an idp or something
        resp = sess.post(self.federated_token_url, authenticated=False)
        return access.create(resp=resp)


class V3FederatedPlugin(utils.TestCase):
    AUTH_URL = 'http://keystone/v3'

    def setUp(self):
        super().setUp()

        self.unscoped_token = fixture.V3Token()
        self.unscoped_token_id = uuid.uuid4().hex
        self.scoped_token = copy.deepcopy(self.unscoped_token)
        self.scoped_token.set_project_scope()
        self.scoped_token.methods.append('token')
        self.scoped_token_id = uuid.uuid4().hex

        s = self.scoped_token.add_service('compute', name='nova')
        s.add_standard_endpoints(
            public='http://nova/public',
            admin='http://nova/admin',
            internal='http://nova/internal',
        )

        self.idp = uuid.uuid4().hex
        self.protocol = uuid.uuid4().hex

        self.token_url = (
            f'{self.AUTH_URL}/OS-FEDERATION/identity_providers/{self.idp}/protocols/{self.protocol}'
            '/auth'
        )

        headers = {'X-Subject-Token': self.unscoped_token_id}
        self.unscoped_mock = self.requests_mock.post(
            self.token_url, json=self.unscoped_token, headers=headers
        )

        headers = {'X-Subject-Token': self.scoped_token_id}
        auth_url = self.AUTH_URL + '/auth/tokens'
        self.scoped_mock = self.requests_mock.post(
            auth_url, json=self.scoped_token, headers=headers
        )

    def get_plugin(self, **kwargs):
        kwargs.setdefault('auth_url', self.AUTH_URL)
        kwargs.setdefault('protocol', self.protocol)
        kwargs.setdefault('identity_provider', self.idp)
        return TesterFederationPlugin(**kwargs)

    def test_federated_url(self):
        plugin = self.get_plugin()
        self.assertEqual(self.token_url, plugin.federated_token_url)

    def test_federated_url_unversioned(self):
        plugin = self.get_plugin(auth_url="http://keystone/")
        self.assertEqual(self.token_url, plugin.federated_token_url)

    def test_unscoped_behaviour(self):
        sess = session.Session(auth=self.get_plugin())
        self.assertEqual(self.unscoped_token_id, sess.get_token())

        self.assertTrue(self.unscoped_mock.called)
        self.assertFalse(self.scoped_mock.called)

    def test_scoped_behaviour(self):
        auth = self.get_plugin(project_id=self.scoped_token.project_id)
        sess = session.Session(auth=auth)
        self.assertEqual(self.scoped_token_id, sess.get_token())

        self.assertTrue(self.unscoped_mock.called)
        self.assertTrue(self.scoped_mock.called)


class K2KAuthPluginTest(utils.TestCase):
    TEST_ROOT_URL = 'http://127.0.0.1:5000/'
    TEST_URL = '{}{}'.format(TEST_ROOT_URL, 'v3')
    TEST_PASS = 'password'

    REQUEST_ECP_URL = TEST_URL + '/auth/OS-FEDERATION/saml2/ecp'

    SP_ROOT_URL = 'https://sp1.com/v3'
    SP_ID = 'sp1'
    SP_URL = 'https://sp1.com/Shibboleth.sso/SAML2/ECP'
    SP_AUTH_URL = (
        SP_ROOT_URL + '/OS-FEDERATION/identity_providers'
        '/testidp/protocols/saml2/auth'
    )

    SERVICE_PROVIDER_DICT = {
        'id': SP_ID,
        'auth_url': SP_AUTH_URL,
        'sp_url': SP_URL,
    }

    def setUp(self):
        super().setUp()
        self.token_v3 = fixture.V3Token()
        self.token_v3.add_service_provider(
            self.SP_ID, self.SP_AUTH_URL, self.SP_URL
        )
        self.session = session.Session()

        self.k2kplugin = self.get_plugin()

    def _get_base_plugin(self):
        self.stub_url(
            'POST',
            ['auth', 'tokens'],
            headers={'X-Subject-Token': uuid.uuid4().hex},
            json=self.token_v3,
        )
        return v3.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )

    def _mock_k2k_flow_urls(self, redirect_code=302):
        # List versions available for auth
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        # The IdP should return a ECP wrapped assertion when requested
        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            content=bytes(k2k_fixtures.ECP_ENVELOPE, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=200,
        )

        # The SP should respond with a redirect (302 or 303)
        self.requests_mock.register_uri(
            'POST',
            self.SP_URL,
            content=bytes(k2k_fixtures.TOKEN_BASED_ECP, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=redirect_code,
        )

        # Should not follow the redirect URL, but use the auth_url attribute
        self.requests_mock.register_uri(
            'GET',
            self.SP_AUTH_URL,
            json=k2k_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': k2k_fixtures.UNSCOPED_TOKEN_HEADER},
        )

    def get_plugin(self, **kwargs):
        kwargs.setdefault('base_plugin', self._get_base_plugin())
        kwargs.setdefault('service_provider', self.SP_ID)
        return v3.Keystone2Keystone(**kwargs)

    def test_remote_url(self):
        remote_auth_url = self.k2kplugin._remote_auth_url(self.SP_AUTH_URL)
        self.assertEqual(self.SP_ROOT_URL, remote_auth_url)

    def test_fail_getting_ecp_assertion(self):
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        self.requests_mock.register_uri(
            'POST', self.REQUEST_ECP_URL, status_code=401
        )

        self.assertRaises(
            exceptions.AuthorizationFailure,
            self.k2kplugin._get_ecp_assertion,
            self.session,
        )

    def test_get_ecp_assertion_empty_response(self):
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            headers={'Content-Type': 'application/vnd.paos+xml'},
            content=b'',
            status_code=200,
        )

        self.assertRaises(
            exceptions.InvalidResponse,
            self.k2kplugin._get_ecp_assertion,
            self.session,
        )

    def test_get_ecp_assertion_wrong_headers(self):
        self.requests_mock.get(
            self.TEST_URL,
            json={'version': fixture.V3Discovery(self.TEST_URL)},
            headers={'Content-Type': 'application/json'},
        )

        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            headers={'Content-Type': uuid.uuid4().hex},
            content=b'',
            status_code=200,
        )

        self.assertRaises(
            exceptions.InvalidResponse,
            self.k2kplugin._get_ecp_assertion,
            self.session,
        )

    def test_send_ecp_authn_response(self):
        self._mock_k2k_flow_urls()
        # Perform the request
        response = self.k2kplugin._send_service_provider_ecp_authn_response(
            self.session, self.SP_URL, self.SP_AUTH_URL
        )

        # Check the response
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER,
            response.headers['X-Subject-Token'],
        )

    def test_send_ecp_authn_response_303_redirect(self):
        self._mock_k2k_flow_urls(redirect_code=303)
        # Perform the request
        response = self.k2kplugin._send_service_provider_ecp_authn_response(
            self.session, self.SP_URL, self.SP_AUTH_URL
        )

        # Check the response
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER,
            response.headers['X-Subject-Token'],
        )

    def test_end_to_end_workflow(self):
        self._mock_k2k_flow_urls()
        auth_ref = self.k2kplugin.get_auth_ref(self.session)
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER, auth_ref.auth_token
        )

    def test_end_to_end_workflow_303_redirect(self):
        self._mock_k2k_flow_urls(redirect_code=303)
        auth_ref = self.k2kplugin.get_auth_ref(self.session)
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER, auth_ref.auth_token
        )

    def test_end_to_end_with_generic_password(self):
        # List versions available for auth
        self.requests_mock.get(
            self.TEST_ROOT_URL,
            json=fixture.DiscoveryList(self.TEST_ROOT_URL),
            headers={'Content-Type': 'application/json'},
        )

        # The IdP should return a ECP wrapped assertion when requested
        self.requests_mock.register_uri(
            'POST',
            self.REQUEST_ECP_URL,
            content=bytes(k2k_fixtures.ECP_ENVELOPE, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=200,
        )

        # The SP should respond with a redirect (302 or 303)
        self.requests_mock.register_uri(
            'POST',
            self.SP_URL,
            content=bytes(k2k_fixtures.TOKEN_BASED_ECP, 'latin-1'),
            headers={'Content-Type': 'application/vnd.paos+xml'},
            status_code=302,
        )

        # Should not follow the redirect URL, but use the auth_url attribute
        self.requests_mock.register_uri(
            'GET',
            self.SP_AUTH_URL,
            json=k2k_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': k2k_fixtures.UNSCOPED_TOKEN_HEADER},
        )

        self.stub_url(
            'POST',
            ['auth', 'tokens'],
            headers={'X-Subject-Token': uuid.uuid4().hex},
            json=self.token_v3,
        )

        plugin = identity.Password(
            self.TEST_ROOT_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            user_domain_id='default',
        )

        k2kplugin = self.get_plugin(base_plugin=plugin)
        self.assertEqual(
            k2k_fixtures.UNSCOPED_TOKEN_HEADER,
            k2kplugin.get_token(self.session),
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

try:
    # explicitly re-export symbol
    # https://mypy.readthedocs.io/en/stable/command_line.html#cmdoption-mypy-no-implicit-reexport
    from lxml import etree as etree
except ImportError:
    etree = None  # type: ignore[assignment]

from keystoneauth1 import exceptions
from keystoneauth1.identity import v3
from keystoneauth1 import session as ks_session


class _Saml2TokenAuthMethod(v3.AuthMethod):
    def __init__(self) -> None: ...

    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: v3.Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> tuple[None, None] | tuple[str, ty.Mapping[str, object]]:
        raise exceptions.HttpNotImplemented(
            'This method should never be called'
        )


_T = ty.TypeVar('_T')


class BaseSAMLPlugin(v3.FederationBaseAuth):
    HTTP_MOVED_TEMPORARILY = 302
    HTTP_SEE_OTHER = 303

    _auth_method_class = _Saml2TokenAuthMethod

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        identity_provider_url: str,
        username: str,
        password: str,
        protocol: str,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ) -> None:
        """Class constructor accepting following parameters.

        :param auth_url: URL of the Identity Service
        :type auth_url: string

        :param identity_provider: Name of the Identity Provider the client
                                  will authenticate against. This parameter
                                  will be used to build a dynamic URL used to
                                  obtain unscoped OpenStack token.
        :type identity_provider: string

        :param identity_provider_url: An Identity Provider URL, where the
                                      SAML2 auhentication request will be
                                      sent.
        :type identity_provider_url: string

        :param username: User's login
        :type username: string

        :param password: User's password
        :type password: string

        :param protocol: Protocol to be used for the authentication.
                         The name must be equal to one configured at the
                         keystone sp side. This value is used for building
                         dynamic authentication URL.
                         Typical value would be: saml2
        :type protocol: string

        """
        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.identity_provider_url = identity_provider_url
        self.username = username
        self.password = password

    @staticmethod
    def _first(_list: list[_T]) -> _T:
        if len(_list) != 1:
            raise IndexError('Only single element list is acceptable')
        return _list[0]

    @staticmethod
    def str_to_xml(
        content: bytes, msg: str | None = None, include_exc: bool = True
    ) -> etree._Element:
        try:
            return etree.XML(content)
        except etree.XMLSyntaxError as e:
            if not msg:
                msg = str(e)
            else:
                msg = msg % e if include_exc else msg
            raise exceptions.AuthorizationFailure(msg)

    @staticmethod
    def xml_to_str(content: etree._Element, **kwargs: ty.Any) -> bytes:
        return ty.cast(bytes, etree.tostring(content, **kwargs))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystoneauth1.extras import _saml2
from keystoneauth1 import loading
from keystoneauth1.loading import opts


class Saml2Password(loading.BaseFederationLoader[_saml2.V3Saml2Password]):
    @property
    def plugin_class(self) -> type[_saml2.V3Saml2Password]:
        return _saml2.V3Saml2Password

    @property
    def available(self) -> bool:
        return _saml2._V3_SAML2_AVAILABLE

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'identity-provider-url',
                    required=True,
                    help=(
                        'An Identity Provider URL, where the SAML2 '
                        'authentication request will be sent.'
                    ),
                ),
                loading.Opt('username', help='Username', required=True),
                loading.Opt(
                    'password', secret=True, help='Password', required=True
                ),
            ]
        )

        return options


class ADFSPassword(loading.BaseFederationLoader[_saml2.V3ADFSPassword]):
    @property
    def plugin_class(self) -> type[_saml2.V3ADFSPassword]:
        return _saml2.V3ADFSPassword

    @property
    def available(self) -> bool:
        return _saml2._V3_ADFS_AVAILABLE

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'identity-provider-url',
                    required=True,
                    help=(
                        'An Identity Provider URL, where the SAML '
                        'authentication request will be sent.'
                    ),
                ),
                loading.Opt(
                    'service-provider-endpoint',
                    required=True,
                    help="Service Provider's Endpoint",
                ),
                loading.Opt(
                    'service-provider-entity-id',
                    required=True,
                    help="Service Provider's SAML Entity ID",
                ),
                loading.Opt('username', help='Username', required=True),
                loading.Opt(
                    'password', secret=True, required=True, help='Password'
                ),
            ]
        )

        return options



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystoneauth1 import loading
from keystoneauth1.loading import opts
from keystoneauth1 import noauth


class NoAuth(loading.BaseLoader[noauth.NoAuth]):
    """Use no tokens to perform requests.

    This can be used to instantiate clients for services deployed in
    noauth/standalone mode.

    There is no fetching a service catalog or determining scope information
    and so it cannot be used by clients that expect to use this scope
    information.

    """

    @property
    def plugin_class(self) -> type[noauth.NoAuth]:
        return noauth.NoAuth

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'endpoint', help='The endpoint that will always be used'
                )
            ]
        )

        return options



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import plugin

if ty.TYPE_CHECKING:
    from keystoneauth1 import session as ks_session

SERVICE_AUTH_HEADER_NAME = 'X-Service-Token'

__all__ = ('ServiceTokenAuthWrapper',)


class ServiceTokenAuthWrapper(plugin.BaseAuthPlugin):
    def __init__(
        self,
        user_auth: plugin.BaseAuthPlugin,
        service_auth: plugin.BaseAuthPlugin,
    ):
        super().__init__()
        self.user_auth = user_auth
        self.service_auth = service_auth

    def get_headers(
        self, session: 'ks_session.Session'
    ) -> dict[str, str] | None:
        headers = self.user_auth.get_headers(session) or {}
        token = self.service_auth.get_token(session)
        if token:
            headers[SERVICE_AUTH_HEADER_NAME] = token

        return headers

    def invalidate(self) -> bool:
        # NOTE(jamielennox): hmm, what to do here? Should we invalidate both
        # the service and user auth? Only one? There's no way to know what the
        # failure was to selectively invalidate.
        user = self.user_auth.invalidate()
        service = self.service_auth.invalidate()
        return user or service

    def get_connection_params(
        self, session: 'ks_session.Session'
    ) -> plugin.ConnectionParams:
        # NOTE(jamielennox): This is also a bit of a guess but unlikely to be a
        # problem in practice. We don't know how merging connection parameters
        # between these plugins will conflict - but there aren't many plugins
        # that set this anyway.
        # Take the service auth params first so that user auth params will be
        # given priority.
        params = self.service_auth.get_connection_params(session)
        params.update(self.user_auth.get_connection_params(session))
        return params

    # TODO(jamielennox): Everything below here is a generic wrapper that could
    # be extracted into a base wrapper class. We can do this as soon as there
    # is a need for it, but we may never actually need it.

    def get_token(self, session: 'ks_session.Session') -> str | None:
        return self.user_auth.get_token(session)

    def get_endpoint(
        self, session: 'ks_session.Session', **kwargs: ty.Any
    ) -> str | None:
        return self.user_auth.get_endpoint(session, **kwargs)

    def get_user_id(self, session: 'ks_session.Session') -> str | None:
        return self.user_auth.get_user_id(session)

    def get_project_id(self, session: 'ks_session.Session') -> str | None:
        return self.user_auth.get_project_id(session)

    def get_sp_auth_url(
        self, session: 'ks_session.Session', sp_id: str
    ) -> str | None:
        return self.user_auth.get_sp_auth_url(session, sp_id)

    def get_sp_url(
        self, session: 'ks_session.Session', sp_id: str
    ) -> str | None:
        return self.user_auth.get_sp_url(session, sp_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import uuid

import requests

from keystoneauth1 import exceptions
from keystoneauth1.extras import _saml2 as saml2
from keystoneauth1 import fixture as ksa_fixtures
from keystoneauth1 import session
from keystoneauth1.tests.unit.extras.saml2 import fixtures as saml2_fixtures
from keystoneauth1.tests.unit.extras.saml2 import utils
from keystoneauth1.tests.unit import matchers

PAOS_HEADER = 'application/vnd.paos+xml'
CONTENT_TYPE_PAOS_HEADER = {'Content-Type': PAOS_HEADER}
InvalidResponse = saml2.v3.saml2.InvalidResponse


class SamlAuth2PluginTests(utils.TestCase):
    """These test ONLY the standalone requests auth plugin.

    Tests for the auth plugin are later so that hopefully these can be
    extracted into it's own module.
    """

    HEADER_MEDIA_TYPE_SEPARATOR = ','

    TEST_USER = 'user'
    TEST_PASS = 'pass'
    TEST_SP_URL = 'http://sp.test'
    TEST_IDP_URL = 'http://idp.test'
    TEST_CONSUMER_URL = "https://openstack4.local/Shibboleth.sso/SAML2/ECP"

    def get_plugin(self, **kwargs):
        kwargs.setdefault('identity_provider_url', self.TEST_IDP_URL)
        kwargs.setdefault('requests_auth', (self.TEST_USER, self.TEST_PASS))
        return saml2.v3.saml2._SamlAuth(**kwargs)

    @property
    def calls(self):
        return [r.url.strip('/') for r in self.requests_mock.request_history]

    def basic_header(self, username=TEST_USER, password=TEST_PASS):
        user_pass = (f'{username}:{password}').encode()
        return 'Basic {}'.format(base64.b64encode(user_pass).decode('utf-8'))

    def test_request_accept_headers(self):
        # Include some random Accept header
        random_header = uuid.uuid4().hex
        headers = {'Accept': random_header}
        req = requests.Request('GET', 'http://another.test', headers=headers)

        plugin = self.get_plugin()
        plugin_headers = plugin(req).headers
        self.assertIn('Accept', plugin_headers)

        # Since we have included a random Accept header, the plugin should have
        # added the PAOS_HEADER to it using the correct media type separator
        accept_header = plugin_headers['Accept']
        self.assertIn(self.HEADER_MEDIA_TYPE_SEPARATOR, accept_header)
        self.assertIn(
            random_header,
            accept_header.split(self.HEADER_MEDIA_TYPE_SEPARATOR),
        )
        self.assertIn(
            PAOS_HEADER, accept_header.split(self.HEADER_MEDIA_TYPE_SEPARATOR)
        )

    def test_passed_when_not_200(self):
        text = uuid.uuid4().hex
        test_url = 'http://another.test'
        self.requests_mock.get(
            test_url,
            status_code=201,
            headers=CONTENT_TYPE_PAOS_HEADER,
            text=text,
        )

        resp = requests.get(test_url, auth=self.get_plugin())
        self.assertEqual(201, resp.status_code)
        self.assertEqual(text, resp.text)

    def test_200_without_paos_header(self):
        text = uuid.uuid4().hex
        test_url = 'http://another.test'
        self.requests_mock.get(test_url, status_code=200, text=text)

        resp = requests.get(test_url, auth=self.get_plugin())
        self.assertEqual(200, resp.status_code)
        self.assertEqual(text, resp.text)

    def test_standard_workflow_302_redirect(self):
        text = uuid.uuid4().hex

        self.requests_mock.get(
            self.TEST_SP_URL,
            response_list=[
                {
                    'headers': CONTENT_TYPE_PAOS_HEADER,
                    'content': utils.make_oneline(
                        saml2_fixtures.SP_SOAP_RESPONSE
                    ),
                },
                {'text': text},
            ],
        )

        authm = self.requests_mock.post(
            self.TEST_IDP_URL, content=saml2_fixtures.SAML2_ASSERTION
        )

        self.requests_mock.post(
            self.TEST_CONSUMER_URL,
            status_code=302,
            headers={'Location': self.TEST_SP_URL},
        )

        resp = requests.get(self.TEST_SP_URL, auth=self.get_plugin())
        self.assertEqual(200, resp.status_code)
        self.assertEqual(text, resp.text)

        self.assertEqual(
            self.calls,
            [
                self.TEST_SP_URL,
                self.TEST_IDP_URL,
                self.TEST_CONSUMER_URL,
                self.TEST_SP_URL,
            ],
        )

        self.assertEqual(
            self.basic_header(), authm.last_request.headers['Authorization']
        )

        authn_request = self.requests_mock.request_history[1].text
        self.assertThat(
            saml2_fixtures.AUTHN_REQUEST, matchers.XMLEquals(authn_request)
        )

    def test_standard_workflow_303_redirect(self):
        text = uuid.uuid4().hex

        self.requests_mock.get(
            self.TEST_SP_URL,
            response_list=[
                {
                    'headers': CONTENT_TYPE_PAOS_HEADER,
                    'content': utils.make_oneline(
                        saml2_fixtures.SP_SOAP_RESPONSE
                    ),
                },
                {'text': text},
            ],
        )

        authm = self.requests_mock.post(
            self.TEST_IDP_URL, content=saml2_fixtures.SAML2_ASSERTION
        )

        self.requests_mock.post(
            self.TEST_CONSUMER_URL,
            status_code=303,
            headers={'Location': self.TEST_SP_URL},
        )

        resp = requests.get(self.TEST_SP_URL, auth=self.get_plugin())
        self.assertEqual(200, resp.status_code)
        self.assertEqual(text, resp.text)

        url_flow = [
            self.TEST_SP_URL,
            self.TEST_IDP_URL,
            self.TEST_CONSUMER_URL,
            self.TEST_SP_URL,
        ]

        self.assertEqual(url_flow, [r.url.rstrip('/') for r in resp.history])
        self.assertEqual(url_flow, self.calls)

        self.assertEqual(
            self.basic_header(), authm.last_request.headers['Authorization']
        )

        authn_request = self.requests_mock.request_history[1].text
        self.assertThat(
            saml2_fixtures.AUTHN_REQUEST, matchers.XMLEquals(authn_request)
        )

    def test_initial_sp_call_invalid_response(self):
        """Send initial SP HTTP request and receive wrong server response."""
        self.requests_mock.get(
            self.TEST_SP_URL,
            headers=CONTENT_TYPE_PAOS_HEADER,
            text='NON XML RESPONSE',
        )

        self.assertRaises(
            InvalidResponse,
            requests.get,
            self.TEST_SP_URL,
            auth=self.get_plugin(),
        )

        self.assertEqual(self.calls, [self.TEST_SP_URL])

    def test_consumer_mismatch_error_workflow(self):
        consumer1 = 'http://consumer1/Shibboleth.sso/SAML2/ECP'
        consumer2 = 'http://consumer2/Shibboleth.sso/SAML2/ECP'
        soap_response = saml2_fixtures.soap_response(consumer=consumer1)
        saml_assertion = saml2_fixtures.saml_assertion(destination=consumer2)

        self.requests_mock.get(
            self.TEST_SP_URL,
            headers=CONTENT_TYPE_PAOS_HEADER,
            content=soap_response,
        )

        self.requests_mock.post(self.TEST_IDP_URL, content=saml_assertion)

        # receive the SAML error, body unchecked
        saml_error = self.requests_mock.post(consumer1)

        self.assertRaises(
            saml2.v3.saml2.ConsumerMismatch,
            requests.get,
            self.TEST_SP_URL,
            auth=self.get_plugin(),
        )

        self.assertTrue(saml_error.called)


class AuthenticateviaSAML2Tests(utils.TestCase):
    TEST_USER = 'user'
    TEST_PASS = 'pass'
    TEST_IDP = 'tester'
    TEST_PROTOCOL = 'saml2'
    TEST_AUTH_URL = 'http://keystone.test:5000/v3/'

    TEST_IDP_URL = 'https://idp.test'
    TEST_CONSUMER_URL = "https://openstack4.local/Shibboleth.sso/SAML2/ECP"

    def get_plugin(self, **kwargs):
        kwargs.setdefault('auth_url', self.TEST_AUTH_URL)
        kwargs.setdefault('username', self.TEST_USER)
        kwargs.setdefault('password', self.TEST_PASS)
        kwargs.setdefault('identity_provider', self.TEST_IDP)
        kwargs.setdefault('identity_provider_url', self.TEST_IDP_URL)
        kwargs.setdefault('protocol', self.TEST_PROTOCOL)
        return saml2.V3Saml2Password(**kwargs)

    def sp_url(self, **kwargs):
        kwargs.setdefault('base', self.TEST_AUTH_URL.rstrip('/'))
        kwargs.setdefault('identity_provider', self.TEST_IDP)
        kwargs.setdefault('protocol', self.TEST_PROTOCOL)

        templ = (
            '%(base)s/OS-FEDERATION/identity_providers/'
            '%(identity_provider)s/protocols/%(protocol)s/auth'
        )
        return templ % kwargs

    @property
    def calls(self):
        return [r.url.strip('/') for r in self.requests_mock.request_history]

    def basic_header(self, username=TEST_USER, password=TEST_PASS):
        user_pass = (f'{username}:{password}').encode()
        return 'Basic {}'.format(base64.b64encode(user_pass).decode('utf-8'))

    def setUp(self):
        super().setUp()
        self.session = session.Session()
        self.default_sp_url = self.sp_url()

    def test_workflow(self):
        token_id = uuid.uuid4().hex
        token = ksa_fixtures.V3Token()

        self.requests_mock.get(
            self.default_sp_url,
            response_list=[
                {
                    'headers': CONTENT_TYPE_PAOS_HEADER,
                    'content': utils.make_oneline(
                        saml2_fixtures.SP_SOAP_RESPONSE
                    ),
                },
                {'headers': {'X-Subject-Token': token_id}, 'json': token},
            ],
        )

        authm = self.requests_mock.post(
            self.TEST_IDP_URL, content=saml2_fixtures.SAML2_ASSERTION
        )

        self.requests_mock.post(
            self.TEST_CONSUMER_URL,
            status_code=302,
            headers={'Location': self.sp_url()},
        )

        auth_ref = self.get_plugin().get_auth_ref(self.session)

        self.assertEqual(token_id, auth_ref.auth_token)

        self.assertEqual(
            self.calls,
            [
                self.default_sp_url,
                self.TEST_IDP_URL,
                self.TEST_CONSUMER_URL,
                self.default_sp_url,
            ],
        )

        self.assertEqual(
            self.basic_header(), authm.last_request.headers['Authorization']
        )

        authn_request = self.requests_mock.request_history[1].text
        self.assertThat(
            saml2_fixtures.AUTHN_REQUEST, matchers.XMLEquals(authn_request)
        )

    def test_consumer_mismatch_error_workflow(self):
        consumer1 = 'http://keystone.test/Shibboleth.sso/SAML2/ECP'
        consumer2 = 'http://consumer2/Shibboleth.sso/SAML2/ECP'

        soap_response = saml2_fixtures.soap_response(consumer=consumer1)
        saml_assertion = saml2_fixtures.saml_assertion(destination=consumer2)

        self.requests_mock.get(
            self.default_sp_url,
            headers=CONTENT_TYPE_PAOS_HEADER,
            content=soap_response,
        )

        self.requests_mock.post(self.TEST_IDP_URL, content=saml_assertion)

        # receive the SAML error, body unchecked
        saml_error = self.requests_mock.post(consumer1)

        self.assertRaises(
            exceptions.AuthorizationFailure,
            self.get_plugin().get_auth_ref,
            self.session,
        )

        self.assertTrue(saml_error.called)

    def test_initial_sp_call_invalid_response(self):
        """Send initial SP HTTP request and receive wrong server response."""
        self.requests_mock.get(
            self.default_sp_url,
            headers=CONTENT_TYPE_PAOS_HEADER,
            text='NON XML RESPONSE',
        )

        self.assertRaises(
            exceptions.AuthorizationFailure,
            self.get_plugin().get_auth_ref,
            self.session,
        )

        self.assertEqual(self.calls, [self.default_sp_url])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1.identity.v3 import base
from keystoneauth1 import session as ks_session


__all__ = ('ReceiptMethod',)


class ReceiptMethod(base.AuthMethod):
    """Construct an Auth plugin to continue authentication with a receipt.

    :param string receipt: Receipt for authentication.
    """

    receipt: str

    def __init__(self, *, receipt: str) -> None:
        self.receipt = receipt

    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: base.Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> tuple[None, None] | tuple[str, ty.Mapping[str, object]]:
        """Add the auth receipt to the headers.

        We explicitly return None to avoid being added to the request
        methods, or body.
        """
        headers['Openstack-Auth-Receipt'] = self.receipt
        return (None, None)

    def get_cache_id_elements(self) -> dict[str, str | None]:
        return {'receipt_receipt': self.receipt}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import fixture as kfixture


def project_scoped_token():
    fixture = kfixture.V3Token(
        user_id='c4da488862bd435c9e6c0275a0d0e49a',
        user_name='exampleuser',
        user_domain_id='4e6893b7ba0b4006840c3845660b86ed',
        user_domain_name='exampledomain',
        expires='2010-11-01T03:32:15-05:00',
        project_id='225da22d3ce34b15877ea70b2a575f58',
        project_name='exampleproject',
        project_domain_id='4e6893b7ba0b4006840c3845660b86ed',
        project_domain_name='exampledomain',
    )

    fixture.add_role(id='76e72a', name='admin')
    fixture.add_role(id='f4f392', name='member')

    region = 'RegionOne'
    tenant = '225da22d3ce34b15877ea70b2a575f58'

    service = fixture.add_service('volume')
    service.add_standard_endpoints(
        public=f'http://public.com:8776/v1/{tenant}',
        internal=f'http://internal:8776/v1/{tenant}',
        admin=f'http://admin:8776/v1/{tenant}',
        region=region,
    )

    service = fixture.add_service('image')
    service.add_standard_endpoints(
        public='http://public.com:9292/v1',
        internal='http://internal:9292/v1',
        admin='http://admin:9292/v1',
        region=region,
    )

    service = fixture.add_service('compute')
    service.add_standard_endpoints(
        public=f'http://public.com:8774/v2/{tenant}',
        internal=f'http://internal:8774/v2/{tenant}',
        admin=f'http://admin:8774/v2/{tenant}',
        region=region,
    )

    service = fixture.add_service('ec2')
    service.add_standard_endpoints(
        public='http://public.com:8773/services/Cloud',
        internal='http://internal:8773/services/Cloud',
        admin='http://admin:8773/services/Admin',
        region=region,
    )

    service = fixture.add_service('identity')
    service.add_standard_endpoints(
        public='http://public.com:5000/v3',
        internal='http://internal:5000/v3',
        admin='http://admin:35357/v3',
        region=region,
    )

    return fixture


def domain_scoped_token():
    fixture = kfixture.V3Token(
        user_id='c4da488862bd435c9e6c0275a0d0e49a',
        user_name='exampleuser',
        user_domain_id='4e6893b7ba0b4006840c3845660b86ed',
        user_domain_name='exampledomain',
        expires='2010-11-01T03:32:15-05:00',
        domain_id='8e9283b7ba0b1038840c3842058b86ab',
        domain_name='anotherdomain',
    )

    fixture.add_role(id='76e72a', name='admin')
    fixture.add_role(id='f4f392', name='member')
    region = 'RegionOne'

    service = fixture.add_service('volume')
    service.add_standard_endpoints(
        public='http://public.com:8776/v1/None',
        internal='http://internal.com:8776/v1/None',
        admin='http://admin.com:8776/v1/None',
        region=region,
    )

    service = fixture.add_service('image')
    service.add_standard_endpoints(
        public='http://public.com:9292/v1',
        internal='http://internal:9292/v1',
        admin='http://admin:9292/v1',
        region=region,
    )

    service = fixture.add_service('compute')
    service.add_standard_endpoints(
        public='http://public.com:8774/v1.1/None',
        internal='http://internal:8774/v1.1/None',
        admin='http://admin:8774/v1.1/None',
        region=region,
    )

    service = fixture.add_service('ec2')
    service.add_standard_endpoints(
        public='http://public.com:8773/services/Cloud',
        internal='http://internal:8773/services/Cloud',
        admin='http://admin:8773/services/Admin',
        region=region,
    )

    service = fixture.add_service('identity')
    service.add_standard_endpoints(
        public='http://public.com:5000/v3',
        internal='http://internal:5000/v3',
        admin='http://admin:35357/v3',
        region=region,
    )

    return fixture


AUTH_SUBJECT_TOKEN = '3e2813b7ba0b4006840c3825860b86ed'

AUTH_RESPONSE_HEADERS = {'X-Subject-Token': AUTH_SUBJECT_TOKEN}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import typing as ty
import uuid

import typing_extensions as ty_ext

from keystoneauth1 import _utils
from keystoneauth1.fixture import exception


class V2Role(ty.TypedDict):
    name: str


class V2Endpoint(ty.TypedDict):
    tenantId: str
    publicURL: str
    adminURL: str
    internalURL: str
    region: str | None
    id: str


class V2Service(ty.TypedDict):
    name: str
    type: str
    endpoints: ty_ext.NotRequired[list[V2Endpoint]]


class V2Tenant(ty.TypedDict, total=False):
    id: str
    name: str


class V2Trust(ty.TypedDict, total=False):
    id: str
    trustee_user_id: str


class V2Token(ty.TypedDict):
    id: str
    expires: str
    issued_at: str
    tenant: ty_ext.NotRequired[V2Tenant]
    audit_ids: ty_ext.NotRequired[list[str]]
    bind: ty_ext.NotRequired[dict[str, ty.Any]]


class V2User(ty.TypedDict):
    id: str
    name: str
    roles: ty_ext.NotRequired[list[V2Role]]


class V2Metadata(ty.TypedDict):
    roles: ty_ext.NotRequired[list[str]]


class V2Access(ty.TypedDict):
    token: V2Token
    user: V2User
    metadata: ty_ext.NotRequired[V2Metadata]
    serviceCatalog: ty_ext.NotRequired[list[dict[str, ty.Any]]]
    trust: ty_ext.NotRequired[V2Trust]


class V2TokenRoot(ty.TypedDict):
    access: V2Access


class _Service(dict[str, ty.Any]):
    def add_endpoint(
        self,
        public: str,
        admin: str | None = None,
        internal: str | None = None,
        tenant_id: str | None = None,
        region: str | None = None,
        id: str | None = None,
    ) -> V2Endpoint:
        data: V2Endpoint = {
            'tenantId': tenant_id or uuid.uuid4().hex,
            'publicURL': public,
            'adminURL': admin or public,
            'internalURL': internal or public,
            'region': region,
            'id': id or uuid.uuid4().hex,
        }

        endpoints = self.setdefault('endpoints', [])
        endpoints.append(data)
        return data


class Token(dict[str, ty.Any]):
    """A V2 Keystone token that can be used for testing.

    This object is designed to allow clients to generate a correct V2 token for
    use in there test code. It should prevent clients from having to know the
    correct token format and allow them to test the portions of token handling
    that matter to them and not copy and paste sample.
    """

    def __init__(
        self,
        token_id: str | None = None,
        expires: datetime.datetime | str | None = None,
        issued: datetime.datetime | str | None = None,
        tenant_id: str | None = None,
        tenant_name: str | None = None,
        user_id: str | None = None,
        user_name: str | None = None,
        trust_id: str | None = None,
        trustee_user_id: str | None = None,
        audit_id: str | None = None,
        audit_chain_id: str | None = None,
    ) -> None:
        super().__init__()

        self.token_id = token_id or uuid.uuid4().hex
        self.user_id = user_id or uuid.uuid4().hex
        self.user_name = user_name or uuid.uuid4().hex
        self.audit_id = audit_id or uuid.uuid4().hex

        if not issued:
            issued = _utils.before_utcnow(minutes=2)

        if not expires and isinstance(issued, datetime.datetime):
            expires = issued + datetime.timedelta(hours=1)

        if isinstance(issued, str):
            self.issued_str = issued
        else:
            self.issued = issued

        if isinstance(expires, str):
            self.expires_str = expires
        elif expires is not None:
            self.expires = expires

        if tenant_id or tenant_name:
            self.set_scope(tenant_id, tenant_name)

        if trust_id or trustee_user_id:
            # the trustee_user_id will generally be the same as the user_id as
            # the token is being issued to the trustee
            self.set_trust(
                id=trust_id, trustee_user_id=trustee_user_id or user_id
            )

        if audit_chain_id:
            self.audit_chain_id = audit_chain_id

    @property
    def root(self) -> V2Access:
        if 'access' not in self:
            self['access'] = {
                'token': {'id': '', 'expires': '', 'issued_at': ''},
                'user': {'id': '', 'name': ''},
            }
        access: V2Access = self['access']
        return access

    @property
    def _token(self) -> V2Token:
        return self.root['token']

    @property
    def token_id(self) -> str:
        return self._token['id']

    @token_id.setter
    def token_id(self, value: str) -> None:
        self._token['id'] = value

    @property
    def expires_str(self) -> str:
        return self._token['expires']

    @expires_str.setter
    def expires_str(self, value: str) -> None:
        self._token['expires'] = value

    @property
    def expires(self) -> datetime.datetime:
        return _utils.parse_isotime(self.expires_str)

    @expires.setter
    def expires(self, value: datetime.datetime) -> None:
        self.expires_str = value.isoformat()

    @property
    def issued_str(self) -> str:
        return self._token['issued_at']

    @issued_str.setter
    def issued_str(self, value: str) -> None:
        self._token['issued_at'] = value

    @property
    def issued(self) -> datetime.datetime:
        return _utils.parse_isotime(self.issued_str)

    @issued.setter
    def issued(self, value: datetime.datetime) -> None:
        self.issued_str = value.isoformat()

    @property
    def _user(self) -> V2User:
        return self.root['user']

    @property
    def user_id(self) -> str:
        return self._user['id']

    @user_id.setter
    def user_id(self, value: str) -> None:
        self._user['id'] = value

    @property
    def user_name(self) -> str:
        return self._user['name']

    @user_name.setter
    def user_name(self, value: str) -> None:
        self._user['name'] = value

    @property
    def tenant_id(self) -> str | None:
        return self._token.get('tenant', {}).get('id')

    @tenant_id.setter
    def tenant_id(self, value: str) -> None:
        if 'tenant' not in self._token:
            self._token['tenant'] = {}
        self._token['tenant']['id'] = value

    @property
    def tenant_name(self) -> str | None:
        return self._token.get('tenant', {}).get('name')

    @tenant_name.setter
    def tenant_name(self, value: str) -> None:
        if 'tenant' not in self._token:
            self._token['tenant'] = {}
        self._token['tenant']['name'] = value

    @property
    def _metadata(self) -> V2Metadata:
        if 'metadata' not in self.root:
            self.root['metadata'] = {}
        return self.root['metadata']

    @property
    def trust_id(self) -> str | None:
        return self.root.get('trust', {}).get('id')

    @trust_id.setter
    def trust_id(self, value: str) -> None:
        if 'trust' not in self.root:
            self.root['trust'] = {}
        self.root['trust']['id'] = value

    @property
    def trustee_user_id(self) -> str | None:
        return self.root.get('trust', {}).get('trustee_user_id')

    @trustee_user_id.setter
    def trustee_user_id(self, value: str) -> None:
        if 'trust' not in self.root:
            self.root['trust'] = {}
        self.root['trust']['trustee_user_id'] = value

    @property
    def audit_id(self) -> str | None:
        try:
            return self._token.get('audit_ids', [])[0]
        except IndexError:
            return None

    @audit_id.setter
    def audit_id(self, value: str) -> None:
        audit_chain_id = self.audit_chain_id
        if audit_chain_id:
            lval = [value, audit_chain_id]
        else:
            lval = [value]
        self._token['audit_ids'] = lval

    @property
    def audit_chain_id(self) -> str | None:
        try:
            return self._token.get('audit_ids', [])[1]
        except IndexError:
            return None

    @audit_chain_id.setter
    def audit_chain_id(self, value: str) -> None:
        audit_id = self.audit_id
        if audit_id:
            self._token['audit_ids'] = [audit_id, value]
        else:
            self._token['audit_ids'] = [value]

    def validate(self) -> None:
        scoped = 'tenant' in self._token
        catalog = self.root.get('serviceCatalog')

        if catalog and not scoped:
            msg = 'You cannot have a service catalog on an unscoped token'
            raise exception.FixtureValidationError(msg)

        if scoped and not self._user.get('roles'):
            msg = 'You must have roles on a token to scope it'
            raise exception.FixtureValidationError(msg)

    def add_role(
        self, name: str | None = None, id: str | None = None
    ) -> dict[str, str]:
        role_id = id or uuid.uuid4().hex
        role_name = name or uuid.uuid4().hex

        if 'roles' not in self._user:
            self._user['roles'] = []
        self._user['roles'].append({'name': role_name})

        if 'roles' not in self._metadata:
            self._metadata['roles'] = []
        self._metadata['roles'].append(role_id)

        return {'id': role_id, 'name': role_name}

    def add_service(self, type: str, name: str | None = None) -> _Service:
        service_name = name or uuid.uuid4().hex
        service = _Service(name=service_name, type=type)

        if 'serviceCatalog' not in self.root:
            self.root['serviceCatalog'] = []
        self.root['serviceCatalog'].append(service)
        return service

    def remove_service(self, type: str) -> None:
        if 'serviceCatalog' in self.root:
            self.root['serviceCatalog'] = [
                f for f in self.root['serviceCatalog'] if f['type'] != type
            ]

    def set_scope(
        self, id: str | None = None, name: str | None = None
    ) -> None:
        self.tenant_id = id or uuid.uuid4().hex
        self.tenant_name = name or uuid.uuid4().hex

    def set_trust(
        self, id: str | None = None, trustee_user_id: str | None = None
    ) -> None:
        self.trust_id = id or uuid.uuid4().hex
        self.trustee_user_id = trustee_user_id or uuid.uuid4().hex

    def set_bind(self, name: str, data: ty.Any) -> None:
        if 'bind' not in self._token:
            self._token['bind'] = {}
        self._token['bind'][name] = data
